import argparse


def main():
    parser = argparse.ArgumentParser(description='Short sample app')
    parser.add_argument('--in1', type=int)
    parser.add_argument('--in2', type=int)
    parser.add_argument('--no_mult', action="store_true", default=False)

    args = parser.parse_args()

    print(args)
    print('in 1  : {}'.format(args.in1))
    print('in 2  : {}'.format(args.in2))
    print('mult: {}'.format(args.no_mult))

    print('ADD=[{}]'.format(args.in1 + args.in2))
    if not args.no_mult:
        print('MULT=[{}]'.format(args.in1 * args.in2))


if __name__ == "__main__":
    main()


#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse


def get_parser():
    parser = argparse.ArgumentParser(description="Run the tests for a Tool to check the function")
    parser.add_argument('tool', metavar='TOOL', help='the id of the tool to test')
    return parser


def main():
    """
    Run the tests of a tool to verify the proper function    
    """
    # No arguments were parsed yet, parse them now
    parser = get_parser()
    args, unknown_args = parser.parse_known_args()

    import fastr
    tool = fastr.toollist[args.tool]
    try:
        tool.test()
    except fastr.exceptions.FastrValueError as e:
        fastr.log.error('Tool is not valid: {}'.format(e))

if __name__ == '__main__':
    main()



import os

from fastr.utils import cmd
from fastr.utils.rest_generation import create_rest_table


def generate_command_reference():
    commands = sorted(cmd.find_commands())

    title = 'Command Line Tools'
    help_file = ['{}\n{}'.format(title, '=' * len(title))]

    links = [":ref:`{c} <cmdline-{c}>`".format(c=c) for c in commands]
    doc_strings = []

    for command in commands:
        module = cmd.get_command_module(command)
        if hasattr(module.main, '__doc__'):
            # Use the first line of the doc string
            doc_strings.append(module.main.__doc__.strip().splitlines()[0])
        else:
            doc_strings.append('')

    help_file.append(create_rest_table(data=[links, doc_strings], headers=['command', 'description']))

    for command in commands:
        heading = '{}\n{}'.format(command, '-' * len(command))

        reference = '.. _cmdline-{}:'.format(command)

        help_text = \
"""
.. argparse::
   :module: fastr.utils.cmd.{c}
   :func: get_parser
   :prog: fastr {c}
""".format(c=command)

        help_file.append('\n\n'.join([
            reference,
            heading,
            help_text
            ])
        )

    filename = os.path.join(os.path.dirname(__file__), 'fastr.commandline.rst')

    with open(filename, 'w') as fh_out:
        fh_out.write('\n\n'.join(help_file))


if __name__ == '__main__':
    generate_command_reference()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imp
import os
import shutil

import fastr


def create_reference(example_path):
    example_base, _ = os.path.splitext(os.path.basename(example_path))
    example_module = imp.load_source(example_base, example_path)

    if hasattr(example_module, 'IS_TEST') and getattr(example_module, 'IS_TEST'):
        network = example_module.create_network()
        output_directory = './data/reference/{n.id}_{n.version}'.format(n=network)

        if os.path.exists(output_directory):
            fastr.log.info('Skipping creation of {}, already exists!'.format(output_directory))
            return

        # Do not publish this to PIM
        fastr.config.pim_host = ''  # Do not publish anything to PIM

        source_data = example_module.source_data(network)
        network.create_reference(source_data=source_data, output_directory=output_directory)
        shutil.rmtree(os.path.join(output_directory, '__fastr_run_tmp__'))


def generate_all_references():
    examples_dir = fastr.config.examplesdir
    examples = [os.path.abspath(os.path.join(examples_dir, x)) for x in os.listdir(examples_dir) if x.endswith('.py')]
    for example in examples:
        create_reference(example)


if __name__ == '__main__':
    generate_all_references()


#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from threading import Timer
from webbrowser import open_new_tab

from fastr import config
from fastr.web.run import runapp


def get_parser():
    parser = argparse.ArgumentParser(description="Fastr web client")
    parser.add_argument('-d', '--debug', action="store_true", help="Debug mode.")
    parser.add_argument('-o', '--openpage', action="store_true", help="Open web page after start.")
    return parser


def open_url(url):
    print('Opening webapp in webbrowser using {}'.format(url))
    open_new_tab(url)


def main():
    """
    Start the fastr webapp and open in a new browser tab
    """
    # No arguments were parsed yet, parse them now
    parser = get_parser()
    args = parser.parse_args()

    if args.openpage:
        url = 'http://{host}:{port}'.format(host=config.web_hostname, port=config.web_port)
        t = Timer(1.0, open_url, args=[url])
        t.daemon = True
        t.start()

    runapp(args.debug)


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import fastr
from fastr.core.samples import SampleIndex, SampleItem
from fastr.plugins import FlowInterface


def kfold(n_items, n_folds):
    items_per_fold = n_items / n_folds
    remaining_items = n_items % n_folds
    fold_sizes = [items_per_fold + 1 if x < remaining_items else items_per_fold for x in range(n_folds)]
    return [(range(0, sum(fold_sizes[:x])) + range(sum(fold_sizes[:x+1]), n_items), range(sum(fold_sizes[:x]), sum(fold_sizes[:x+1]))) for x in range(n_folds)]


class CrossValidation(FlowInterface.flow_plugin_type):
    """
    Advanced flow plugin that generated a cross-validation data flow. The node
    need an input with data and an input number of folds. Based on that the
    outputs test and train will be supplied with a number of data sets.
    """
    @staticmethod
    def execute(payload):
        log_data = None

        items = payload['items']
        method = sum(x.data for x in payload['method']).sequence_part()
        number_of_folds = sum(x.data for x in payload['number_of_folds']).sequence_part()
        labels = None

        if len(method) != 1:
            raise ValueError('Can only handle 1 method for cross validation!')
        method = method[0].value

        if len(number_of_folds) != 1:
            raise ValueError('Can only handle 1 number_of_folds for cross validation!')
        number_of_folds = number_of_folds[0].value

        fastr.log.debug('CV Plugin items: {!r}'.format(items))
        fastr.log.info('CV Plugin method: {!r}'.format(method))
        fastr.log.info('CV Plugin number_of_folds: {!r}'.format(number_of_folds))

        if labels is not None and len(labels) != len(items):
            raise ValueError('If given, the number of labels should match the number of items!')

        if method == 'KFold':
            cv_iterator = kfold(len(items), n_folds=number_of_folds)
        else:
            raise ValueError('Invalid method selected!')

        train_data = {}
        test_data = {}

        for fold, (train, test) in enumerate(cv_iterator):
            fold_id = 'fold_{}'.format(fold)

            for k, index in enumerate(train):
                item = items[index]
                new_index = SampleIndex([k, fold])
                new_id = item.id + fold_id

                train_data[new_index, new_id] = SampleItem(new_index, new_id, item.data, item.jobs)

            for k, index in enumerate(test):
                item = items[index]
                new_index = SampleIndex([k, fold])
                new_id = item.id + fold_id

                test_data[new_index, new_id] = SampleItem(new_index, new_id, item.data, item.jobs)

        result_data = {'train': train_data, 'test': test_data}

        for key, value in result_data.items():
            values = [(x.id, x.index, x.data) for x in value.values()]
            fastr.log.debug('Result data {}: {}'.format(key, values))

        return result_data, log_data



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict, Mapping

# Import similar to nipype
try:
    import nipype
    import traits as traits
    IMPORT_SUCCESS = True
    IMPORT_ERROR = ''
except ImportError as e:
    IMPORT_SUCCESS = False
    IMPORT_ERROR = e.message

import fastr
from fastr import exceptions
from fastr.core.baseplugin import PluginState
from fastr.core.interface import Interface, InterfaceResult, InputSpec, OutputSpec


class HiddenFieldMap(Mapping):
    def __init__(self, *args, **kwargs):
        self._data = OrderedDict(*args, **kwargs)

    def __getitem__(self, item):
        return self._data[item]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        for key, value in self._data.items():
            if not (hasattr(value, 'hidden') and value.hidden):
                yield key


if IMPORT_SUCCESS:
    DATATYPE_MAP = {
        traits.trait_types.Int: 'Int',
        traits.trait_types.Long: 'Int',
        traits.trait_types.Float: 'Float',
        traits.trait_types.Bool: 'Boolean',
        traits.trait_types.Str: 'String',
        traits.trait_types.String: 'String',
        traits.trait_types.Unicode: 'String',
        traits.trait_types.List: None,
    }
else:
    DATATYPE_MAP = {}


class NipypeInterface(Interface):
    """
    Experimental interfaces to using ``nipype`` interfaces directly in fastr
    tools, only using a simple reference.

    To create a tool using a nipype interface just create an interface with
    the correct type and set the ``nipype`` argument to the correct class.
    For example in an xml tool this would become::

        <interface class="NipypeInterface">
          <nipype_class>nipype.interfaces.elastix.Registration</nipype_class>
        </interface>

    .. note::

        To use these interfaces ``nipype`` should be installed on the system.

    .. warning::

        This interface plugin is basically functional, but highly experimental!
    """

    if IMPORT_ERROR:
        _status = (PluginState.failed, 'nipype could not be found!')

    def __init__(self, id_, nipype_cls=None, document=None):
        super(NipypeInterface, self).__init__()

        if nipype_cls is None and document is None:
            raise exceptions.FastrValueError('Need to have at Nipype class set in either the nipype_cls or document argument')
        elif nipype_cls is not None and document is not None:
            raise exceptions.FastrValueError('Need to have at Nipype class set in either the nipype_cls or the document argument, not both!')
        elif document is not None:
            nipype_cls = document['nipype_class']

        if isinstance(nipype_cls, (str, unicode)):
            if '.' in nipype_cls:
                module, cls = nipype_cls.rsplit('.', 1)
                _temp = __import__(module, globals(), locals(), [cls], -1)
                nipype_cls = getattr(_temp, cls)
            else:
                raise ValueError('If a string is given, it should be a full class specification (in the form of nipype.interfaces.something.Class), got "{}"'.format(nipype_cls))

        self.nipype_class = nipype_cls

        #: The ID of the interface
        self.id = id_

        # Create the inputs and outputs based on the nipype class
        self._create_inputs_outputs()

    def _create_inputs_outputs(self):
        # Try to extract the correct input output descriptions
        nipype_object = self.nipype_class()
        ignore_inputs = ['args', 'environ', 'ignore_exception', 'terminal_output']

        inputs = []
        for id_, input_ in nipype_object.input_spec().items():
            default = input_.default
            if default == '':
                default = None

            inputs.append(InputSpec(id_=id_,
                                    cardinality=1 if not input_.is_trait_type(traits.trait_types.List) else '1-*',
                                    datatype=self.get_type(input_),
                                    required=input_.mandatory and default is None,
                                    description=input_.desc,
                                    default=default,
                                    hidden=id_ in ignore_inputs))

        outputs = []
        for id_, output in nipype_object.output_spec().items():
            outputs.append(OutputSpec(id_=id_,
                                      cardinality=1 if not output.is_trait_type(traits.trait_types.List) else 'unknown',
                                      datatype=self.get_type(output),
                                      automatic=True,
                                      required=False,
                                      description=output.desc,
                                      hidden=False))

        # Create the inputs/outputs spec to expose to the rest of the system
        self._inputs = HiddenFieldMap((input_.id, input_) for input_ in inputs)
        self._outputs = HiddenFieldMap((output_.id, output_) for output_ in outputs)

    def __getstate__(self):
        state = {
            'id': self.id,
            'class': type(self).__name__,
            'nipype_class': self.nipype_class
        }

        return state

    def __setstate__(self, state):
        del state['class']
        self.__dict__.update(state)

        nipype_cls = self.nipype_class
        if isinstance(nipype_cls, (str, unicode)):
            if '.' in nipype_cls:
                module, cls = nipype_cls.rsplit('.', 1)
                _temp = __import__(module, globals(), locals(), [cls], -1)
                nipype_cls = getattr(_temp, cls)
            else:
                raise ValueError('If a string is given, it should be a full class specification (in the form of nipype.interfaces.something.Class), got "{}"'.format(nipype_cls))

        self.nipype_class = nipype_cls

        self._create_inputs_outputs()

    def __eq__(self, other):
        if not isinstance(other, NipypeInterface):
            return NotImplemented

        return vars(self) == vars(other)

    def get_type(self, trait):
        datatype = DATATYPE_MAP.get(trait.trait_type, 'AnyFile')
        if datatype is None:
            new_trait = trait.inner_traits[0]
            return self.get_type(new_trait)
        else:
            return datatype

    @property
    def inputs(self):
        return self._inputs

    @property
    def outputs(self):
        return self._outputs

    @property
    def expanding(self):
        return 0

    def execute(self, target, payload):
        """
        Execute the interface using a specific target and payload (containing
        a set of values for the arguments)

        :param target: the target to use
        :type target: :py:class:`SampleId <fastr.core.target.Target>`
        :param dict payload: the values for the arguments
        :return: result of the execution
        :rtype: InterfaceResult
        """
        # Create the nipype interface object
        nipype_object = self.nipype_class()

        # Set all the inputs
        for id_, input_value in payload['inputs'].items():
            if len(input_value) == 0:
                fastr.log.info('Skipping non-valued input {}'.format(id_))
                continue

            value = [x.value if isinstance(x, fastr.datatypes.DataType) else x for x in input_value]
            if self.inputs[id_].cardinality == 1:
                if len(value) > 1:
                    raise exceptions.FastrValueError('Expected cardinality 1, got {}'.format(len(value)))
                value = value[0]
            else:
                value = list(value)

            setattr(nipype_object.inputs, id_, value)

        # Run the interface
        nipype_result = nipype_object.run()

        result_data = nipype_result.outputs.get()

        fastr.log.info('NIPYPE RESULT: {}'.format(result_data))

        # Parse into correct format
        for key, value in result_data.items():
            if isinstance(value, list):
                value = tuple(value)
            else:
                value = (value,)

            result_data[key] = value

        fastr.log.info('RESULT DATA: {}'.format(result_data))
        # Wrap result in InterfaceResult
        result = InterfaceResult(result_data, nipype_result.runtime.dictcopy(), payload)
        return result

    @classmethod
    def test(cls):
        if not IMPORT_SUCCESS:
            raise ImportError('Could not load required module: {}'.format(IMPORT_ERROR))



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import imp

import nose.tools as nt

import fastr


def test_examples():
    examples_dir = fastr.config.examplesdir
    examples = [os.path.join(examples_dir, x) for x in os.listdir(examples_dir) if x.endswith(".py")]
    for example_path in examples:
        example_base, _ = os.path.splitext(os.path.basename(example_path))
        try:
            example_module = imp.load_source(example_base, example_path)
        except:
            # Invalid module, assume that is should be ignored
            continue

        # Only use tests that are marked as such
        if hasattr(example_module, 'IS_TEST') and getattr(example_module, 'IS_TEST'):
            yield run, example_path


def run(example_path):
    # Set some config variables
    fastr.config.pim_host = ''  # Do not publish anything to PIM
    fastr.config.execution_plugin = "BlockingExecution"  # This is the safest option

    example_base, _ = os.path.splitext(os.path.basename(example_path))
    example_module = imp.load_source(example_base, example_path)
    network = example_module.create_network()

    reference_dir = os.path.join(os.path.dirname(example_path),
                                 'data',
                                 'reference',
                                 '{}_{}'.format(network.id, network.version))

    fastr.log.info('Using reference dir: {}'.format(reference_dir))
    if os.path.isdir(reference_dir):
        result = network.test(reference_dir)
        nt.ok_(result, 'Test Network run did not yield same result as the reference run (see log)')
    else:
        nt.ok_(False, 'Could not find reference data for test!')




import argparse


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', metavar='N',
                        dest='input', type=float, required=True, nargs='+',
                        help='Values to perform max on')
    args = parser.parse_args()

    print('Maximum = {}'.format(max(args.input)))


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The base class for all Plugins in the fastr system
"""

from abc import ABCMeta
from enum import Enum
import gzip
import StringIO

import fastr
from colorama import Fore, Back, Style

from fastr import exceptions
from fastr.utils.classproperty import classproperty


class PluginState(Enum):
    uninitialized = Back.CYAN + 'UnInitialized' + Style.RESET_ALL
    preload = Back.LIGHTGREEN_EX + 'PreLoad' + Style.RESET_ALL
    unloaded = Back.CYAN + 'UnLoaded' + Style.RESET_ALL
    failed = Fore.WHITE + Back.RED + Style.BRIGHT + 'Failed' + Style.RESET_ALL
    loaded = Fore.WHITE + Back.GREEN + Style.BRIGHT + 'Loaded' + Style.RESET_ALL


class PluginMeta(ABCMeta):
    def __repr__(self):
        return '<{}: {} class [{}]>'.format(self.__bases__[0].__name__, self.id, getattr(self.status, 'value', 'NA'))


class BasePlugin(object):
    """
    Base class for Plugins in the fastr system.
    """
    __metaclass__ = PluginMeta

    #: The status of the plugin
    _status = (PluginState.uninitialized, 'Plugin object created', None)
    _source_code = None
    module = None
    _instantiate = False

    def __init__(self):
        """
        The BasePlugin constructor.

        :return: the created plugin
        :rtype: BasePlugin
        :raises FastrPluginNotLoaded: if the plugin did not load correctly
        """

        if self.status not in (PluginState.preload, PluginState.loaded):
            raise exceptions.FastrPluginNotLoaded('Plugin was not properly loaded: [{}] {}'.format(self._status[0], self._status[1]))

    def __str__(self):
        """
        Creare string representation of the plugin.

        :return: string represenation
        :rtype: str
        """
        return '<Plugin: {}>'.format(self.__class__.__name__)

    def __repr__(self):
        return '<{}: {} object [{}]>'.format(type(self).__bases__[0].__name__, self.id, getattr(self.status, 'value', 'NA'))

    @classproperty
    def id(cls):
        """
        The id of this plugin
        """
        return cls.__name__

    @classproperty
    def configuration_fields(cls):
        """
        Plugins can register the fields they need from the configuration
        by adding them here. The format of the data structure is::

            {
                "var_name": (type, default, description),
            }

        So for example::

            {
                "some_string": (str, "spam", "some string used for demonstration purposes"),
                "some_int": (str, 42, "the answer to life etc"),
            }
        """
        return {}

    @classmethod
    def register_configuration(cls):
        """
        Register and test the configuation fields of the plugin
        """
        # Register fields
        fastr.config.register_fields(cls.configuration_fields)

        # Validate that current values are correct
        for field_name, field_spec in cls.configuration_fields.items():
            fastr.log.debug('Tested config value: {}'.format(getattr(fastr.config, field_name)))

    @classproperty
    def fullid(cls):
        """
        The full id of this plugin
        """
        return 'fastr://plugins/{}'.format(cls.id)

    @classproperty
    def source_code(cls):
        """
        The source code of this plugin
        """
        code_gzip = StringIO.StringIO(cls._source_code)
        with gzip.GzipFile(fileobj=code_gzip, mode='rb') as gzip_stream:
            return gzip_stream.read()

    @classproperty
    def status(cls):
        """
        The status of the plugin.
        """
        return cls._status[0]

    @classproperty
    def instantiate(cls):
        """
        Flag if the plugin should be instantiated
        """
        return cls._instantiate

    @classproperty
    def status_message(cls):
        """
        The message explaining the status of the plugin.
        """
        return cls._status[1]

    @classmethod
    def set_status(cls, status, message, exception=None):
        """
        Update the status of the plugin

        :param str status: the new status
        :param str message: message explaining the status change
        :param str exception: stacktrace of the exception causing the failed load
        """
        if not isinstance(status, PluginState):
            raise exceptions.FastrTypeError('Plugin status should be of type PluginState')

        if not isinstance(message, basestring):
            raise exceptions.FastrTypeError('Plugin status message should be a string')

        cls._status = (status, message, exception)

    @classmethod
    def set_code(cls, source_code):
        """
        Set the filename and source code of the plugin

        :param str source_code: the source code of the plugin
        """
        #if cls._source_code is not None:
        #    raise exceptions.FastrCannotChangeAttributeError('Filename and/or source code are already set!')

        # We gzip the source code before storing
        #code_gzip = StringIO.StringIO()
        #with gzip.GzipFile(fileobj=code_gzip, mode='w') as gzip_stream:
        #    gzip_stream.write(source_code)
        #cls._source_code = code_gzip.getvalue()
        cls._source_code = source_code

    def cleanup(self):
        """
        Perform any cleanup action needed when the plugin use ended. This can
        be closing files/streams etc.
        """
        pass

    @classmethod
    def test(cls):
        """
        Test the plugin, default behaviour is just to instantiate the plugin
        """
        obj = cls()
        obj.cleanup()
        del obj


class Plugin(BasePlugin):
    pass



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import fastr.execution.executionscript


def get_parser():
    parser = argparse.ArgumentParser(description="Execute a job or network from commandline")
    parser.add_argument('network', metavar='NETWORKFILE', type=str, help='File of the network to execute')
    return parser


def create_network_parser(network):
    parser = argparse.ArgumentParser(description="Execute the {} Network".format(network.id))
    for source in network.sourcelist:
        parser.add_argument('--{}'.format(source),
                            metavar=source.upper(),
                            required=True,
                            nargs='*',
                            help='Source data for {}'.format(source))

    for sink in network.sinklist:
        parser.add_argument('--{}'.format(sink),
                            metavar=sink.upper(),
                            required=True,
                            help='The Sink output specification for {}'.format(sink))
    return parser


def main():
    """
    Run a Network from the commandline
    """
    # No arguments were parsed yet, parse them now
    parser = get_parser()
    args = parser.parse_args()

    network = fastr.Network.loadf(args.network)
    network_parser = create_network_parser(network)
    network_args = vars(network_parser.parse_args(unknown_args))

    # Split data in Source and Sink data
    source_data = {}
    for source in network.sourcelist:
        source_data[source] = network_args[source]

    sink_data = {}
    for sink in network.sinklist:
        sink_data[sink] = network_args[sink]

    fastr.log.info('Source data: {}'.format(source_data))
    fastr.log.info('Sink data: {}'.format(sink_data))

    network.execute(source_data, sink_data)


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from collections import namedtuple, OrderedDict, Mapping
import os
import sys
import re

import fastr
from fastr import exceptions
from fastr.core.interface import Interface, InterfaceResult, InputSpec, OutputSpec
from fastr.core.baseplugin import Plugin
from fastr.core.pluginmanager import PluginSubManager
from fastr.core.serializable import Serializable
from fastr.datatypes import EnumType, TypeGroup, URLType


class PathDescription(object):
    def __init__(self, path):
        """
        Create a path description that allows access to parts

        :param str path: the path to describe
        """
        if not isinstance(path, basestring):
            path = str(path)

        self.path = path

    def __repr__(self):
        return '<PathSpec: {}>'.format(self.path)

    def __str__(self):
        return self.path

    def __getitem__(self, item):
        """
        Forward the getitem to the path string

        :param int,slice item: the part of the path the retrieve
        :return: substring
        """
        return self.path[item]

    @property
    def directory(self):
        """
        The directory name of the path
        """
        return os.path.dirname(self.path)

    @property
    def filename(self):
        """
        The entire filename of the path
        """
        return os.path.basename(self.path)

    @property
    def basename(self):
        """
        The basename of the filename of the path
        """
        return self.filename.split('.', 1)[0]

    @property
    def extension(self):
        """
        The extension of the path
        """
        if '.' not in self.path:
            return ''
        else:
            return self.path.split('.', 1)[1]


class CollectorPlugin(Plugin):
    """
    :py:class:`CollectorPlugins <fastr.resources.plugins.interfaceplugins.fasterinterface.CollectorPlugin>`
    are used for finding and collecting the output data of outputs part of a
    :py:class:`FastrInterface <fastr.resources.plugins.interfaceplugins.fasterinterface.FasterInterface>`
    """

    # Signal managers that this plugin should be stored instantiated
    _instantiate = True

    def __init__(self):
        """
        Constructor
        """
        super(CollectorPlugin, self).__init__()
        self.job = None

    @property
    def fullid(self):
        """
        The full id of the plugin
        """
        if self.job is not None:
            return 'fastr://plugins/collectorplugins/{}/{}'.format(self.id, self.job.id)
        else:
            return 'fastr://plugins/collectorplugins/{}'.format(self.id)

    def collect_results(self, interface, output, result):
        """
        Start the collection of the results. This method calls the actual
        implementation from the subclass (_collect_results) and wraps it
        with some convience functionality.

        :param interface: Job to collect data for
        :param output: Output to collect data for
        """
        self._collect_results(interface, output, result)

    @abstractmethod
    def _collect_results(self, job, output, result):
        """
        Placeholder method for the actual collection of the results. This
        method should implemented by subclasses.

        :param job: Job to collect data for
        :param output: Output to collect data for
        """
        pass

    def _regexp_path(self, path):
        """
        Helper function that searches for a regular expression in a path. There
        can be wildcards in any level of the path.

        :param path: path with regular expressions
        :return: list of paths that match the path pattern
        """
        # Get a clean, absolute path to work with
        path = os.path.expanduser(path)
        path = os.path.abspath(path)

        subpaths = self._full_split(path)
        if subpaths[0] != os.path.sep and not re.match(r'[a-zA-Z]:[/\\]', subpaths[0]):
            raise ValueError('path should always contain an absolute path (subpaths: {})'.format(subpaths))

        basepath = subpaths[0]

        pathlist = [basepath]
        fastr.log.debug('Basepath: {}'.format(basepath))
        for subpath in subpaths[1:]:
            subpath = '^' + subpath + '$'
            # Test the regexp and give a more understandable error if it fails
            try:
                re.compile(subpath)
            except Exception as detail:
                raise ValueError('Error parsing regexp "{}": {}'.format(subpath, detail))

            # Scan new level for matches
            newpathlist = []
            for curpath in pathlist:
                contents = os.listdir(curpath)
                for option in contents:
                    if re.match(subpath, option):
                        newpathlist.append(os.path.join(curpath, option))

            pathlist = newpathlist

        return pathlist

    @staticmethod
    def _full_split(urlpath):
        """
        Split an urls path completely into parts

        :param urlpath: path part of the url
        :return: a list of parts
        """
        parts = []
        while True:
            newpath, tail = os.path.split(urlpath)
            if newpath == urlpath:
                assert not tail
                if urlpath:
                    parts.append(urlpath)
                break
            parts.append(tail)
            urlpath = newpath
        parts.reverse()
        return parts


class CollectorPluginManager(PluginSubManager):
    """
    Container holding all the CollectorPlugins
    """

    def __init__(self):
        """
        Create the Coll
        :param path:
        :param recursive:
        :return:
        """
        super(CollectorPluginManager, self).__init__(parent=fastr.plugin_manager,
                                                     plugin_class=CollectorPlugin)
        self._key_map = {}

    @property
    def _instantiate(self):
        """
        Indicate that the plugins should instantiated before stored
        """
        return True

    def __keytransform__(self, key):
        try:
            return self._key_map[key]
        except KeyError:
            self._key_map.clear()
            for id_, value in self.data.items():
                self._key_map[value.id] = id_
            return self._key_map[key]

    def __iter__(self):
        for value in self.data.values():
            yield value.id


class HiddenFieldMap(Mapping):
    def __init__(self, *args, **kwargs):
        self._data = OrderedDict(*args, **kwargs)

    def __getitem__(self, item):
        return self._data[item]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        for key, value in self._data.items():
            if not (hasattr(value, 'hidden') and value.hidden):
                yield key


class FastrInterface(Interface):
    """
    The default Interface for fastr. For the command-line Tools as used by
    fastr.
    """

    __dataschemafile__ = 'FastrInterface.schema.json'

    collectors = CollectorPluginManager()
    collector_plugin_type = CollectorPlugin

    def __init__(self, id_, document):
        super(FastrInterface, self).__init__()

        # Load from file if it is not a dict
        if not isinstance(document, dict):
            fastr.log.debug('Trying to load file: {}'.format(document))
            filename = os.path.expanduser(document)
            filename = os.path.abspath(filename)
            document = self._loadf(filename)
        else:
            document = self.get_serializer().instantiate(document)

        #: The ID of the interface
        self.id = id_

        #: List of input parameter descriptions
        self.input_map = OrderedDict()

        #: List of output parameter descriptions
        self.output_map = OrderedDict()

        # Parse input and output fields into parameter_description objects
        for n_order, input_ in enumerate(document['inputs']):
            self.input_map[input_['id']] = InputParameterDescription(self, input_, n_order)
        n_inputs = len(self.input_map)

        for n_order, output in enumerate(document['outputs']):
            self.output_map[output['id']] = OutputParameterDescription(self, output, n_order + n_inputs)

        # Create the inputs/outputs spec to expose to the rest of the system
        self._inputs = HiddenFieldMap((k, InputSpec(id_=v.id,
                                                    cardinality=v.cardinality,
                                                    datatype=v.datatype,
                                                    required=v.required,
                                                    description=v.description,
                                                    default=v.default,
                                                    hidden=v.hidden)) for k, v in self.input_map.items())

        self._outputs = HiddenFieldMap((k, OutputSpec(id_=v.id,
                                                      cardinality=v.cardinality,
                                                      datatype=v.datatype,
                                                      automatic=v.automatic,
                                                      required=v.required,
                                                      description=v.description,
                                                      hidden=v.hidden)) for k, v in self.output_map.items())

    def __eq__(self, other):
        if not isinstance(other, FastrInterface):
            return NotImplemented

        return vars(self) == vars(other)

    def __getstate__(self):
        """
        Get the state of the FastrInterface object.

        :return: state of interface
        :rtype: dict
        """
        state = {
            'id': self.id,
            'class': type(self).__name__,
            'inputs': [x.__getstate__() for x in self.input_map.values()],
            'outputs': [x.__getstate__() for x in self.output_map.values()],
        }

        return state

    def __setstate__(self, state):
        """
        Set the state of the Interface
        """
        self.id = state['id']

        self.input_map = OrderedDict()
        self.output_map = OrderedDict()

        state['inputs'] = state['inputs'] or {}
        for order, x in enumerate(state['inputs']):
            self.input_map[x['id']] = InputParameterDescription(self, x, order)
        state['outputs'] = state['outputs'] or {}
        for order, x in enumerate(state['outputs']):
            self.output_map[x['id']] = OutputParameterDescription(self, x, order)

        # Create the inputs/outputs spec to expose to the rest of the system
        self._inputs = HiddenFieldMap((k, InputSpec(id_=v.id,
                                                    cardinality=v.cardinality,
                                                    datatype=v.datatype,
                                                    required=v.required,
                                                    description=v.description,
                                                    default=v.default,
                                                    hidden=v.hidden)) for k, v in self.input_map.items())

        self._outputs = HiddenFieldMap((k, OutputSpec(id_=v.id,
                                                      cardinality=v.cardinality,
                                                      datatype=v.datatype,
                                                      automatic=v.automatic,
                                                      required=v.required,
                                                      description=v.description,
                                                      hidden=v.hidden)) for k, v in self.output_map.items())

    @property
    def inputs(self):
        return self._inputs

    @property
    def outputs(self):
        return self._outputs

    @property
    def expanding(self):
        return 0

    def execute(self, target, payload):
        """
        Execute the interface using a specific target and payload (containing
        a set of values for the arguments)

        :param target: the target to use
        :type target: :py:class:`SampleId <fastr.core.target.Target>`
        :param dict payload: the values for the arguments
        :return: result of the execution
        :rtype: InterfaceResult
        """
        fastr.log.info('Execution payload: {}'.format(payload))
        command = [target.binary] + self.get_arguments(payload)
        log_data = target.run_command(command)

        result = InterfaceResult(result_data={}, log_data=log_data, payload=payload, errors=None)

        # TODO: add the collection of results and log store here
        fastr.log.info('Collecting results')
        for output in self.outputs.values():
            if not output.automatic:
                if output.id in payload['outputs']:
                    result.result_data[output.id] = payload['outputs'][output.id]
                elif output.required:
                    raise exceptions.FastrValueError('Required output {} not in payload!'.format(output.id))

        self.collect_results(result)

        return result

    def check_input_id(self, id_):
        """
        Check if an id for an object is valid and unused in the Tool. The
        method will always returns True if it does not raise an exception.

        :param str id_: the id to check
        :return: True
        :raises FastrValueError: if the id is not correctly formatted
        :raises FastrValueError: if the id is already in use
        """

        regex = r'^\w[\w\d_]*$'
        if re.match(regex, id_) is None:
            raise exceptions.FastrValueError('An id in Fastr should follow'
                                             ' the following pattern {}'
                                             ' (found {})'.format(regex, id_))

        if id_ in self.input_map:
            raise exceptions.FastrValueError('The id {} is already in use in {}!'.format(id_, self.id))

        return True

    def check_output_id(self, id_):
        """
        Check if an id for an object is valid and unused in the Tool. The
        method will always returns True if it does not raise an exception.

        :param str id_: the id to check
        :return: True
        :raises FastrValueError: if the id is not correctly formatted
        :raises FastrValueError: if the id is already in use
        """

        regex = r'^\w[\w\d_]*$'
        if re.match(regex, id_) is None:
            raise exceptions.FastrValueError('An id in Fastr should follow'
                                             ' the following pattern {}'
                                             ' (found {})'.format(regex, id_))

        if id_ in self.output_map:
            raise exceptions.FastrValueError('The id {} is already in use in {}!'.format(id_, self.id))

        return True

    def get_arguments(self, values):
        """
        Get the argument list for this interface

        :return: return list of arguments
        """
        # Get the argument list
        arguments = self.input_map.values() + self.output_map.values()
        arguments = sorted(arguments, key=lambda x: x.order if x.order >= 0 else sys.maxsize - x.order)

        argument_list = []

        for argument in arguments:
            id_ = argument.id

            try:
                if isinstance(argument, InputParameterDescription):
                    value = values['inputs'][id_]
                else:
                    value = values['outputs'][id_]

                # TODO: check if this works!
                #if len(value) != argument.cardinality:
                #    raise exceptions.FastrValueError('Cardinality of argument {} has an incorrect cardinality (expected {} ({}), found {})'.format(argument.id,
                #                                                                                                                                   argument.cardinality,
                #                                                                                                                                   type(argument.cardinality).__name__,
                #                                                                                                                                   len(value)))
            except KeyError:
                if argument.default is not None:
                    if isinstance(argument.default, list):
                        value = tuple(argument.default)
                    else:
                        value = (argument.default,)
                elif argument.required and not (isinstance(argument, OutputParameterDescription) and argument.automatic):
                    raise exceptions.FastrValueError('Required argument "{}" has no value!'.format(argument.id))
                else:
                    value = None

            if value is not None:
                fastr.log.info('Adding {} to argument list based on {}'.format(value, argument))
                argument_list.extend(argument.get_commandline_argument(value))

        return argument_list

    def collect_results(self, result):
        """
        Collect all results of the interface
        """
        for output in self.output_map.values():
            if output.automatic:
                collector = self.collectors[output.method]

                # Do the try/except here so one failed output collection
                # does not stop the rest
                try:
                    collector.collect_results(self, output, result)

                    fastr.log.debug('Collected automatic result for {}'.format(output.id))
                    fastr.log.debug(result.result_data[output.id])
                except exceptions.FastrError as exception:
                    if output.required:
                        fastr.log.error('Encountered an error when collecting the results')
                        result.errors.append(exception.excerpt())
                    else:
                        fastr.log.info(('Non-required output {} did not yield'
                                        ' results due to collection error: {}').format(output.id,
                                                                                       exception))

    def get_specials(self, payload, output, cardinality_nr):
        """
        Get special attributes. Returns tuples for specials, inputs and outputs
        that are used for formatting substitutions.

        :param output: Output for which to get the specials
        :param int cardinality_nr: the cardinality number
        """
        if output is not None:
            datatype = fastr.typelist[output.datatype]
            # Get extension match
            if issubclass(datatype, URLType):
                extension = datatype.extension
            elif issubclass(datatype, TypeGroup):
                extensions = set()
                for member in datatype.members:
                    extensions.add(member.extension if member.extension is not None else '')
                extension = '({})'.format('|'.join(extensions))
            else:
                extension = ''
        else:
            extension = ''

        specials_tuple_type = namedtuple('Specials', ['cardinality', 'extension'])
        specials = specials_tuple_type(cardinality_nr, extension)

        if len(self.input_map) > 0:
            # Gather the data
            input_data = OrderedDict()
            for input_id, input_ in self.input_map.items():
                temp_data = payload['inputs'][input_id] if input_id in payload['inputs'] else [input_.default]
                input_data[input_id] = [PathDescription(x) for x in temp_data]

            # Create a named tuple
            inputs_tuple_type = namedtuple('Inputs', input_data.keys())
            inputs = inputs_tuple_type(*input_data.values())
        else:
            inputs = ()

        output_arguments = [x for x in self.output_map.values() if not x.automatic]
        if len(output_arguments) > 0:
            # Gather the data
            output_data = OrderedDict()
            for output_id, ouput in self.output_map.items():
                temp_data = payload['outputs'][output_id] if output_id in payload['outputs'] else [output.default]
                output_data[output_id] = [PathDescription(x) for x in temp_data]

            # Create a named tuple
            outputs_tuple_type = namedtuple('Outputs', [x.id for x in self.output_map.values()])
            outputs = outputs_tuple_type(*output_data.values())
        else:
            outputs = ()

        return specials, inputs, outputs


class ParameterDescription(Serializable):
    """
    Description of an input or output parameter used by a Tool. This is the
    super class for both input and output, containing the shared parts.
    """
    _IS_INPUT = False

    def __init__(self, parent, element, order=0):
        """
        Instantiate a ParameterDescription

        :param Tool parent: parent tool
        :param dict element: description of the parameter
        :param order: the order in which the parameter was defined in the file
        """
        self.parent = parent

        if isinstance(element, dict):
            self.id = element['id']
            self.name = element['name']

            if 'datatype' in element:
                self.datatype = element['datatype']
            elif 'enum' in element:
                element['enum'] = [str(x) for x in element['enum']]
                self.datatype = '__{}__{}__Enum__'.format(parent.id, self.id)
                fastr.typelist.create_enumtype(self.datatype, tuple(element['enum']), self.datatype)
            else:
                raise exceptions.FastrValueError('No valid datatype defined for {}'.format(self.id))

            self.prefix = element.get('prefix', None)
            self.repeat_prefix = element.get('repeat_prefix', False)
            self.cardinality = element.get('cardinality', 1)
            self.nospace = element.get('nospace', False)
            self.format = element.get('format', None)
            self.required = element.get('required', True)
            self.default = element.get('default', None)
            self.description = element.get('description', '')
            self.order = element.get('order', order)
            self.hidden = element.get('hidden', False)
        else:
            raise exceptions.FastrTypeError('element should be a dict')

    def __eq__(self, other):
        """
        Compare two ParameterDescription instance with eachother. This
        function helps ignores the parent, but once tests the values for
        equality

        :param other: the other instances to compare to

        :returns: True if equal, False otherwise
        """
        if not isinstance(other, ParameterDescription):
            return NotImplemented

        dict_self = {k: v for k, v in self.__dict__.items()}
        del dict_self['parent']

        dict_other = {k: v for k, v in other.__dict__.items()}
        del dict_other['parent']

        return dict_self == dict_other

    def __getstate__(self):
        """
        Retrieve the state of the ParameterDescription

        :return: the state of the object
        :rtype dict:
        """
        state = {k: v for k, v in self.__dict__.items()}
        state['parent'] = self.parent.id

        datatype = fastr.typelist[self.datatype]
        if issubclass(datatype, EnumType):
            del state['datatype']
            state['enum'] = list(datatype.options)

        return state

    def __setstate__(self, state):
        """
        Set the state of the ParameterDescription by the given state.

        :param dict state: The state to populate the object with
        """
        if 'datatype' not in state:
            if 'enum' in state:
                typename = '__{}__{}__Enum__'.format(state['parent'], state['id'])
                state['datatype'] = typename
                fastr.typelist.create_enumtype(typename, tuple(state['enum']), typename)
            else:
                raise exceptions.FastrValueError('No valid datatype defined for {} in {}'.format(state['id'], state))

        self.__dict__.update(state)

    def get_commandline_argument(self, value):
        """
        Get the commandline argument for this Parameter given the values
        assigned to it.

        :param value: the value(s) for this input
        :return: commandline arguments
        :rtype: list
        """
        argument = []
        if isinstance(value, tuple):
            for cardinality_nr, value_item in enumerate(value):
                value_item = self.format_argument_value(value_item)
                argument.extend(self.format_prefix(value_item, cardinality_nr))

        elif isinstance(value, OrderedDict):
            for cardinality_nr, (key, value_item) in enumerate(value.items()):
                value_item = ','.join(self.format_argument_value(x) for x in value_item if x is not None)
                argument.extend(self.format_prefix('{}={}'.format(key, value_item), cardinality_nr))
        else:
            raise exceptions.FastrTypeError('Argument should be tuple or OrderedDict!')

        return argument

    def format_prefix(self, value, cardinality_nr):
        extra_argument = []

        # Build the prefix
        prefix = self.prefix
        if prefix is not None:
            if prefix.strip() == '':
                prefix = None
            elif cardinality_nr == 0 or self.repeat_prefix:
                prefix = self.prefix.replace('#', str(cardinality_nr))
            else:
                prefix = None

        # Handle special case for boolean flags
        if value.startswith('__FASTR_FLAG__'):
            if value == '__FASTR_FLAG__FALSE__':
                return []
            else:
                if prefix is None:
                    raise exceptions.FastrValueError('Found non-prefixed arguments with type'
                                                     ' Boolean (flag). This is not possible,'
                                                     ' a flag should always have a proper prefix!')
                return [self.prefix.replace('#', str(cardinality_nr))]

        # Handle the normal cases
        if self.nospace:
            if prefix is None:
                raise exceptions.FastrValueError('Found a nospace without a valid prefix! When'
                                                 ' nospace is true a valid prefix should be'
                                                 ' specified (and repeat prefix should be used)!')
            extra_argument.append('{}{}'.format(self.prefix, value))
        else:
            if prefix is not None:
                extra_argument.append(prefix)
            if value is not None:
                extra_argument.append(value)

        return extra_argument

    def format_argument_value(self, value, datatype=None):
        if datatype is None:
            datatype = fastr.typelist[self.datatype]

        if not datatype.isinstance(value):
            fastr.log.debug('CREATING DATATYPE {!r} for {!r}!'.format(datatype, value))
            value = datatype(value)

        # Format (and validate if required) the input value
        value.format = self.format
        if self._IS_INPUT and not value.valid:
            fastr.log.debug('SELF TYPE: {}, ID {}'.format(type(self).__name__, self.id))
            raise exceptions.FastrDataTypeValueError('Value for input {} not a valid instance of type {} (value: {} ({} / {!r}) -> {})'.format(self.id, datatype.id, value, type(value).__name__, value.value, value.valid))

        value = str(value)

        # Filter out boolean flags
        return value


class InputParameterDescription(ParameterDescription):
    """
    Description of an input parameter used by a Tool.
    """
    _IS_INPUT = True

    def __init__(self, parent, element, order=0):
        """
        Instantiate an InputParameterDescription

        :param Tool parent: parent tool
        :param dict element: description of the parameter
        :param order: the order in which the parameter was defined in the file
        """
        if isinstance(element, dict):
            super(InputParameterDescription, self).__init__(parent, element, order)
            self.parent.check_input_id(element['id'])
        else:
            raise exceptions.FastrTypeError('element should be a dict')


class OutputParameterDescription(ParameterDescription):
    """
    Description of a output parameter used by a Tool.
    """
    def __init__(self, parent, element, order=0):
        """
        Instantiate an OutputParameterDescription

        :param Tool parent: parent tool
        :param dict element: description of the parameter
        :param order: the order in which the parameter was defined in the file
        """
        if isinstance(element, dict):
            super(OutputParameterDescription, self).__init__(parent, element, order)
            self.parent.check_output_id(element['id'])
            self.automatic = element.get('automatic', False)
            if self.automatic and self.prefix is not None:
                # Automatic outputs with a prefix are not required by default
                self.required = element.get('required', False)
            self.action = element.get('action', None)
            self.location = element.get('location', None)
            self.separator = element.get('separator', None)
            if self.automatic:
                self.method = element.get('method', 'path')
            else:
                self.method = element.get('method', None)
            self.negate = element.get('negate', False)
        else:
            raise exceptions.FastrTypeError('element should be a dict')

    def __setstate__(self, state):
        """
        Set the state of the OutputParameterDescription by the given state.

        :param dict state: The state to populate the object with
        """
        if 'location' not in state:
            state['location'] = None

        if 'separator' not in state:
            state['separator'] = None

        if 'method' not in state:
            if 'automatic' not in state or not state['automatic']:
                state['method'] = None
            else:
                state['method'] = 'path'

        if 'action' not in state:
            state['action'] = None

        if 'negate' not in state:
            state['negate'] = False

        super(OutputParameterDescription, self).__setstate__(state)

    def get_commandline_argument(self, values):
        """
        Get the commandline argument for this Parameter given the values
        assigned to it.

        :param value: the value(s) for this input
        :return: commandline arguments
        :rtype: list
        """
        if self.action is not None:
            if self.action in self.ACTIONS:
                action_func = self.ACTIONS[self.action]
                for value in values:
                    fastr.log.info('Calling action {} for {}'.format(action_func, value))
                    action_func(self, value)
            else:
                fastr.log.warning('Encountered unknown action {}, only know: {}'.format(self.action,
                                                                                        self.ACTIONS.keys()))

        if not self.automatic or self.prefix is not None:
            if self.automatic and self.prefix and self.negate:
                fastr.log.info('Negating value for automatic prefixed output: {}'.format(self.id))
                fastr.log.info('Values before: {}'.format(values))
                if isinstance(values, tuple):
                    values = tuple(fastr.datatypes.Boolean(not x.value) for x in values)
                elif isinstance(value, OrderedDict):
                    for key, value in values.items():
                        values[key] = tuple(fastr.datatypes.Boolean(not x.value) for x in value)
                else:
                    raise exceptions.FastrTypeError('Argument should be tuple or OrderedDict!')
                fastr.log.info('Values after: {}'.format(values))

            return super(OutputParameterDescription, self).get_commandline_argument(values)
        else:
            return []

    def format_argument_value(self, value):
        # For the non-automatic case, nothing special happens
        if self.automatic:
            datatype = fastr.typelist['Boolean']
            return super(OutputParameterDescription, self).format_argument_value(value,
                                                                                 datatype=datatype)
        else:
            return super(OutputParameterDescription, self).format_argument_value(value)

    def mkdirs(self, value):
        """
        Create the directory if it does not yet exist

        :param URLType value: the directory to create
        """
        value = str(value)
        if not os.path.exists(value):
            os.makedirs(value)

    ACTIONS = {'ensure': mkdirs,
               'mkdir': mkdirs}



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains all Fastr-related Exceptions
"""

import inspect
import os
import textwrap

# pylint: disable=too-many-ancestors
# Because fo inheriting from FastrError and a common exception causes this
# exception, even though this behaviour is desired


class FastrError(Exception):
    """
    This is the base class for all fastr related exceptions. Catching this
    class of exceptions should ensure a proper execution of fastr.
    """
    def __init__(self, *args, **kwargs):
        """
        Constructor for all exceptions. Saves the caller object fullid (if
        found) and the file, function and line number where the object was
        created.
        """
        super(FastrError, self).__init__(*args, **kwargs)

        frame = inspect.stack()[1][0]
        call_object = frame.f_locals.get('self', None)
        if call_object is not None and hasattr(call_object, 'fullid'):
            self.fastr_object = call_object.fullid
        else:
            self.fastr_object = None

        info = inspect.getframeinfo(frame)
        self.filename = info.filename
        self.function = info.function
        self.linenumber = info.lineno

    def __str__(self):
        """
        String representation of the error

        :return: error string
        :rtype: str
        """
        if self.fastr_object is not None:
            return '[{}] {}'.format(self.fastr_object, super(FastrError, self).__str__())
        else:
            return super(FastrError, self).__str__()

    def excerpt(self):
        """
        Return a excerpt of the Error as a tuple.
        """
        return type(self).__name__, self.message, self.filename, self.linenumber


class FastrNotImplementedError(FastrError, NotImplementedError):
    """
    This function/method has not been implemented on purpose (e.g. should be
    overwritten in a sub-class)
    """
    pass


class FastrOptionalModuleNotAvailableError(FastrNotImplementedError):
    """
    A optional modules for Fastr is needed for this function, but is not
    available on the current python installation.
    """
    pass


class FastrValueError(FastrError, ValueError):
    """
    ValueError in the fastr system
    """
    pass


class FastrCannotChangeAttributeError(FastrError):
    """
    Attempting to change an attribute of an object that can be set only once.
    """


class FastrTypeError(FastrError, TypeError):
    """
    TypeError in the fastr system
    """
    pass


class FastrKeyError(FastrError, KeyError):
    """
    KeyError in the fastr system
    """
    pass


class FastrIndexError(FastrError, IndexError):
    """
    IndexError in the fastr system
    """
    pass


class FastrIndexNonexistent(FastrIndexError):
    """
    This is an IndexError for samples requested from a sparse data array. The
    sample is not there but is probably not there because of sparseness rather
    than being a missing sample (e.g. out of bounds).
    """
    pass


class FastrAttributeError(FastrError, AttributeError):
    """
    AttributeError in the fastr system
    """
    pass


class FastrIOError(FastrError, IOError):
    """
    IOError in the fastr system
    """
    pass


class FastrOSError(FastrError, OSError):
    """
    OSError in the fastr system
    """
    pass


class FastrImportError(FastrError, ImportError):
    """
    ImportError in the fastr system
    """
    pass


class FastrStateError(FastrError):
    """
    An object is in an invalid/unexpected state.
    """
    pass


class FastrSizeUnknownError(FastrError):
    """
    The size of object is not (yet) known and only a theoretical estimate is
    available at the moment.
    """
    pass


class FastrSizeMismatchError(FastrError):
    """
    The size of two object in fastr is not matching where it should.
    """
    pass


class FastrSizeInvalidError(FastrError):
    """
    The given size cannot be valid.
    """
    pass


class FastrCardinalityError(FastrError):
    """
    The description of the cardinality is not valid.
    """
    pass


class FastrUnknownURLSchemeError(FastrKeyError):
    """
    Fastr encountered a data URL with a scheme that was not recognised
    by the IOPlugin manager.
    """
    pass


class FastrDataTypeFileNotReadable(FastrError):
    """
    Could not read the datatype file.
    """
    pass


class FastrDataTypeNotAvailableError(FastrError):
    """
    The DataType requested is not found by the fastr system. Typically this
    means that no matching DataType is found in the DataTypeManager.
    """
    pass


class FastrDataTypeNotInstantiableError(FastrError):
    """
    The base classes for DataTypes cannot be instantiated and should always
    be sub-classed.
    """
    pass


class FastrDataTypeMismatchError(FastrError):
    """
    When using a DataType as the key for the DataTypeManager, the
    DataTypeManager found another DataType with the same name already
    in the DataTypeManager. The means fastr has two version of the same
    DataType in the system, which should never happen!
    """
    pass


class FastrDataTypeValueError(FastrError):
    """
    This value in fastr did not pass the validation specificied for its
    DataType, typically means that the data is missing or corrupt.
    """
    pass


class FastrToolNotAvailableError(FastrError):
    """
    The tool used is not available on the current platform (OS and architecture
     combination) and cannot be used.
    """
    pass


class FastrToolTargetNotFound(FastrError):
    """
    Could not determine the location of the tools target binary/script. The
    tool cannot be used.
    """
    pass


class FastrObjectUnknownError(FastrKeyError):
    """
    Reference to a Tool that is not recognised by the fastr system. This
    typically means the specific id/version combination of the requested
    tool has not been loaded by the ToolManager.
    """
    pass


class FastrNetworkUnknownError(FastrKeyError):
    """
    Reference to a Tool that is not recognised by the fastr system. This
    typically means the specific id/version combination of the requested
    tool has not been loaded by the ToolManager.
    """
    pass


class FastrToolUnknownError(FastrKeyError):
    """
    Reference to a Tool that is not recognised by the fastr system. This
    typically means the specific id/version combination of the requested
    tool has not been loaded by the ToolManager.
    """
    pass


class FastrPluginNotAvailable(FastrKeyError):
    """
    Indicates that a requested Plugin was not found on the system.
    """
    pass

class FastrPluginNotLoaded(FastrStateError):
    """
    The plugin was not successfully loaded. This means the plugin class cannot
    be instantiated.
    """
    pass


class FastrNodeNotValidError(FastrStateError):
    """
    A Node is not in a valid state where it should be, typically an invalid
    Node is passed to the executor causing trouble.
    """
    pass


class FastrNodeNotPreparedError(FastrStateError):
    """
    When trying to access executation data of a Node, the Node must be prepare.
    The Node has not been prepared by the execution, so the data is not
    available!
    """
    pass


class FastrNodeAreadyPreparedError(FastrStateError):
    """
    A attempt is made at preparing a Node for the second time. This is not
    allowed as it would wipe the current execution data and cause data-loss.
    """
    pass


class FastrLookupError(FastrError):
    """
    Could not find specified object in the fastr environment.
    """
    pass


class FastrNetworkMismatchError(FastrError):
    """
    Two interacting objects belong to different fastr network.
    """
    pass


class FastrParentMismatchError(FastrError):
    """
    Two interactive objects have different parent where they should be the same
    """
    pass


class FastrSerializationMethodError(FastrKeyError):
    """
    The desired serialization method does not exist.
    """
    pass


class FastrSerializationError(FastrError):
    """
    The serialization encountered a serious problem
    """

    def __init__(self, message, serializer, original_exception=None):
        # Call superclass
        super(FastrSerializationError, self).__init__(message)

        self.data_path = serializer.data_path
        self.schema_path = serializer.schema_path
        self.object_type = serializer.current_type
        self.current_object = serializer.current_object
        self.current_schema = serializer.current_schema
        if original_exception is not None:
            self.original_exception = original_exception
        else:
            self.original_exception = ''

    def __repr__(self):
        """
        Simple string representation of the exception
        """
        return '<{}: {}>'.format(self.__class__.__name__, self.message)

    def __str__(self):
        """
        Advanced string representation of the exception including the data
        about where in the schema things went wrong.
        """
        return self.message + textwrap.dedent('''

            DATA PATH: {}
            SCHEMA PATH: {}
            CURRENT OBJECT TYPE: {}
            CURRENT OBJECT: {}
            CURRENT SCHEMA: {}

            {}
            '''.rstrip()).format(self.data_path,
                                 self.schema_path,
                                 self.object_type,
                                 self.current_object,
                                 self.current_schema,
                                 self.original_exception)


class FastrSerializationIgnoreDefaultError(FastrSerializationError):
    """
    The value and default are both None, so the value should not be serialized.
    """
    pass


class FastrSerializationInvalidDataError(FastrSerializationError):
    """
    Encountered data to serialize that is invalid given the serialization
    schema.
    """
    pass


class FastrVersionInvalidError(FastrValueError):
    """
    The string representation of the version is malformatted.
    """
    pass


class FastrVersionMismatchError(FastrValueError):
    """
    There is a mismatch between different parts of the Fastr environment
    and integrity is compromised.
    """
    pass


class FastrMountUnknownError(FastrKeyError):
    """
    Trying to access an undefined mount
    """
    pass


class FastrSourceDataUnavailableError(FastrKeyError):
    """
    Could not find the Source data for the desire source.
    """
    pass


class FastrSinkDataUnavailableError(FastrKeyError):
    """
    Could not find the Sink data for the desire sink.
    """
    pass


class FastrExecutionError(FastrError):
    """
    Base class for all fastr execution related errors
    """
    pass


class FastrExecutableNotFoundError(FastrExecutionError):
    """
    The executable could not be found!
    """
    def __init__(self, executable=None, *args, **kwargs):
        super(FastrExecutableNotFoundError, self).__init__(*args, **kwargs)
        self.executable = executable
        self.path = os.environ.get('PATH', '').split(os.pathsep)

    def __str__(self):
        """String representation of the error"""
        return 'Could not find executable "{}" on PATH: {}'.format(self.executable, self.path)


class FastrNotExecutableError(FastrExecutionError):
    """
    The command invoked by subprocess is not executable on the system
    """
    pass


class FastrErrorInSubprocess(FastrExecutionError):
    """
    Encountered an error in the subprocess started by the execution script
    """
    pass


class FastrSubprocessNotFinished(FastrExecutionError):
    """
    Encountered an error before the subprocess call by the execution script was
    properly finished.
    """
    pass


class FastrResultFileNotFound(FastrExecutionError):
    """
    Could not found the result file of job that finished. This means the
    executionscript process was killed during interruption. Generally this
    means a scheduler killed it because of resource shortage.
    """
    pass


class FastrOutputValidationError(FastrExecutionError):
    """
    An output of a Job does not pass validation
    """
    pass


class FastrNoValidTargetError(FastrKeyError):
    """
    Cannot find a valid target for the tool
    """
    pass


class FastrCollectorError(FastrError):
    """
    Cannot collect the results from a Job because of an error
    """
    pass


class FastrPluginCapabilityNotImplemented(FastrNotImplementedError):
    """
    A plugin did not implement a capability that it advertised.
    """


# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nose.tools as nt

import fastr
from fastr.core.network import Network
from fastr.utils.dicteq import diffobj_str


class TestNetwork():

    def setup(self):
        self.populated_network = fastr.Network('populated_network')

        self.source_node1 = fastr.SourceNode('Int', 'source')
        self.constant_node = fastr.ConstantNode('Int', (42, 7, 1337), 'constant')

        self.addint = fastr.Node(fastr.toollist['AddInt'], 'testnode_addint')
        self.div_link = self.populated_network.create_link(self.constant_node.output, self.addint.inputs['left_hand'])
        self.div_link.expand = True
        self.addint.inputs['right_hand'] = self.source_node1.output

        self.sink_node = fastr.SinkNode('Int', 'sink')
        self.sink_node.inputs['input'] = self.addint.outputs['result']

        self.empty_network = fastr.Network('empty_network')

    def test_network_eq_operator(self):
        nt.eq_(self.populated_network, self.populated_network)
        nt.eq_(self.empty_network, self.empty_network)
        nt.ok_(self.populated_network != self.empty_network)

    def test_network_getstate_setstate(self):
        #  Dump state and rebuild in a clean object
        empty_state = self.empty_network.__getstate__()
        empty_rebuilt = Network.__new__(Network)
        empty_rebuilt.__setstate__(empty_state)
        nt.eq_(self.empty_network, empty_rebuilt, msg=diffobj_str(self.empty_network, empty_rebuilt))

        #  Dump state and rebuild in a clean object
        populated_state = self.populated_network.__getstate__()
        populated_rebuilt = Network.__new__(Network)
        populated_rebuilt.__setstate__(populated_state)
        nt.eq_(self.populated_network, populated_rebuilt, msg=diffobj_str(self.populated_network, populated_rebuilt))

    def test_network_serializing(self):
        empty_json = self.empty_network.dumps()
        empty_rebuilt = Network.loads(empty_json)
        nt.eq_(self.empty_network, empty_rebuilt)

        populated_json = self.populated_network.dumps()
        populated_rebuilt = Network.loads(populated_json)
        nt.eq_(self.populated_network, populated_rebuilt)

        nt.ok_(populated_rebuilt, empty_rebuilt)
        nt.ok_(self.empty_network != populated_rebuilt)
        nt.ok_(self.populated_network != empty_rebuilt)





#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A small tool to wipe all .pyc files from fastr
"""
import os


def dir_list(directory):
    """
    Find all .pyc files

    :param str directory: directory to search
    :return: all .pyc files
    :rtype: list
    """
    output_list = []
    for root, _, files in os.walk(directory):
        for filename in files:
            if filename.endswith('.pyc'):
                output_list.append(os.path.abspath(os.path.join(root, filename)))
    return output_list


def main():
    """
    Main entry poitn
    """
    os.chdir(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
    print("************* REMOVING .pyc FILES ***********************")
    for filename in dir_list('.'):
        os.remove(filename)
        print('removed {}'.format(filename))
    print("************* ALL .pyc FILES REMOVED ********************")


if __name__ == '__main__':
    main()



import datetime
import functools
import multiprocessing
import threading

import fastr
from fastr.web import app
from flask import url_for, abort
from flask_restplus import Api, Resource, reqparse, fields

api = Api(app,
          doc='/api/doc/',
          version='0.1',
          title='Fastr REST API',
          description='A REST API to interact with Fastr via the web',
          default_mediatype='application/json')

runs = {}


STATUS_MANAGER = multiprocessing.Manager()


def update_status(job, job_status):
    job_status[job.id] = str(job.status), str(job.node_id)


def update_job_result(job, job_status, job_results):
    job_results[job.id] = str(job.output_data)


def network_lock_thread(lock, network):
    fastr.log.debug('WAITING FOR LOCK')
    with lock:
        fastr.log.debug('CALLING NETWORK ABORT')
        network.abort()


def network_runner(network, source_data, sink_data, chuck_status, job_status, job_results, abort_lock):
    network.job_finished_callback = functools.partial(update_job_result, job_status=job_status, job_results=job_results)
    network.job_status_callback = functools.partial(update_status, job_status=job_status)
    abort_lock.acquire()
    abort_thread = threading.Thread(name="NetworkAbort", target=network_lock_thread, args=(abort_lock, network))
    abort_thread.start()

    network.execute(source_data, sink_data)


class Run(object):
    def __init__(self, id_, network, source_data, sink_data):
        self.id = id_
        self.chunks = STATUS_MANAGER.list()
        self.jobs = STATUS_MANAGER.dict()
        self.job_results = STATUS_MANAGER.dict()
        self.source_data = source_data
        self.sink_data = sink_data
        self.network = network.id
        self.abort_lock = multiprocessing.Lock()
        self.process = self.run_network(network, source_data, sink_data, self.abort_lock)

    def run_network(self, network, source_data, sink_data, abort_lock):
        process = multiprocessing.Process(target=network_runner,
                                          args=(network,
                                                source_data,
                                                sink_data,
                                                self.chunks,
                                                self.jobs,
                                                self.job_results,
                                                abort_lock),
                                          name=self.id)
        process.start()
        return process

    def status(self):
        return {'job_status': dict(self.jobs),
                'job_results': dict(self.job_results)}

    def abort(self):
        fastr.log.debug('RELEASING ABORT LOCK')
        self.abort_lock.release()

        if self.process:
            fastr.log.debug('JOINING PROCESS')
            self.process.join(timeout=3)

            if self.process.is_alive():
                fastr.log.debug('TERMINATING PROCESS')
                self.process.terminate()


class ObjectUrl(fields.Raw):
    __schema_type__ = "string"

    def __init__(self, object_classs, **kwargs):
        super(ObjectUrl, self).__init__(**kwargs)
        self._object_class = object_classs

    def format(self, value):
        if isinstance(self._object_class, str):
            return url_for(self._object_class, id=value)
        else:
            return api.url_for(self._object_class, id=value)


class SubUrl(fields.Raw):
    __schema_type__ = "string"

    def __init__(self, object_classs, subfield, **kwargs):
        super(SubUrl, self).__init__(**kwargs)
        self._object_class = object_classs
        self._subfield = subfield

    def format(self, value):
        if isinstance(self._object_class, str):
            url = url_for(self._object_class, id=value)
        else:
            url = api.url_for(self._object_class, id=value)

        return '{}/{}'.format(url, self._subfield)


class ToolApi(Resource):
    @api.response(200, 'Success')
    @api.response(404, 'Tool not found')
    def get(self, id_, version=None):
        """
        Get a Tool json description from the server
        """
        try:
            return fastr.toollist[id_, version].dumps(method='json')
        except KeyError:
            abort(404)


tool_list_model = api.model('ToolList', {
    'tools': fields.List(fields.String),
})


class ToolListApi(Resource):
    @api.marshal_with(tool_list_model)
    def get(self):
        """
        Get a list of all Tools known to the server
        """
        data = {'tools': [url_for('api_version_tool', id=x.ns_id, version=x.command_version) for x in fastr.toollist.values()]}
        print('Data {}'.format(data))
        return data


class NetworkApi(Resource):
    @api.response(200, 'Success')
    @api.response(404, 'Network not found')
    def get(self, id_):
        """
        Get a Network json description from the server
        """
        try:
            return fastr.networklist[id_].dumps(method='dict')
        except KeyError:
            abort(404)


network_list_model = api.model('NetworkList', {
    'networks': fields.List(ObjectUrl('api_network', attribute='id'))
})


class NetworkListApi(Resource):
    @api.marshal_with(network_list_model)
    def get(self):
        """
        Get a list of the networks
        """
        data = {'networks': fastr.networklist.values()}
        print('Data: {}'.format(data))
        return data


run_model = api.model('Run', {
    'url': fields.Url,
    'network': ObjectUrl('api_network', attribute='network'),
    'status': ObjectUrl('api_status', attribute='id'),
    'source_data': fields.Raw,
    'sink_data': fields.Raw,
})


class RunApi(Resource):
    """
    Run API documentation
    """
    @api.response(200, 'Success')
    @api.response(404, 'Network not found')
    @api.marshal_with(run_model)
    def get(self, id_):
        """
        Get information about a Network run
        """
        try:
            return runs[id_]
        except KeyError:
            abort(404)

    @api.response(204, 'Aborted Network run')
    @api.response(404, 'Network not found')
    def delete(self, id_):
        """
        Abort a Network run and stop all associated execution
        """
        if id_ in runs:
            runs[id_].abort()
            return None, 204
        else:
            return None, 404


run_list_model = api.model("RunList", {
    'runs': fields.List(ObjectUrl('api_run', attribute='id'))
})


class RunListApi(Resource):
    request_parser = reqparse.RequestParser()
    request_parser.add_argument('network', type=str, required=True, location='json',
                                help='No network id specified')
    request_parser.add_argument('source_data', type=dict, required=True, location='json',
                                help='No source data was supplied')
    request_parser.add_argument('sink_data', type=dict, required=True, location='json',
                                help='No sink data was supplied')

    @api.marshal_with(run_list_model)
    def get(self):
        """
        Get a list of all Network runs on the server
        """
        return {'runs': runs.values()}

    @api.expect(request_parser)
    @api.response(201, "Created Network run")
    def post(self):
        """
        Create a new Network run and start execution
        """
        args = self.request_parser.parse_args()

        network = fastr.networklist[args['network']]
        run_id = '{}_{}'.format(network.id, datetime.datetime.now().isoformat())
        runs[run_id] = Run(run_id, network, args['source_data'], args['sink_data'])

        return {'run_id': run_id,
                'run': url_for('api_run', id=run_id, _external=True),
                'status': url_for('api_status', id=run_id, _external=True)}, 201, {'Location': url_for('api_run', id=run_id)}


class StatusApi(Resource):
    @api.response(200, "Success")
    @api.response(404, "Run not found")
    def get(self, id_):
        """
        Get the status of a Network Run on the server
        """
        try:
            return runs[id_].status()
        except KeyError:
            abort(404)


api.add_resource(NetworkApi, '/api/networks/<id>', endpoint='api_network')
api.add_resource(NetworkListApi, '/api/networks', endpoint='api_networks')
api.add_resource(ToolApi, '/api/tools/<id>', endpoint='api_tool')
api.add_resource(ToolApi, '/api/tools/<id>/<version>', endpoint='api_version_tool')
api.add_resource(ToolListApi, '/api/tools', endpoint='api_tools')
api.add_resource(RunApi, '/api/runs/<id>', endpoint='api_run')
api.add_resource(RunListApi, '/api/runs', endpoint='api_runs')
api.add_resource(StatusApi, '/api/runs/<id>/status', endpoint='api_status')




# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the NetworkChunker class and its default implementation
the DefaultNetworkChunker
"""

from abc import abstractmethod
from collections import deque
import fastr
import fastr.exceptions as exceptions


class NetworkChunker(object):
    """
    The base class for NetworkChunkers. A Network chunker is a class that takes
    a Network and produces a list of chunks that can each be analyzed and
    executed in one go.
    """

    @abstractmethod
    def chunck_network(self, network):
        """ Create a list of Network chunks that can be pre-analyzed completely.
        Each chunk needs to be executed before the next can be analyzed and
        executed.

        :param network: Network to split into chunks
        :return: list containing chunks
        """
        pass


class DefaultNetworkChunker(NetworkChunker):
    """
    The default implementation of the NetworkChunker. It tries to create as
    large as possible chunks so the execution blocks as little as possible.
    """

    def __init__(self):
        self.chunks = []
        self.node_status = {}
        self.pool = set()
        self.processed_nodes = set()
        self.used_nodes = set()

    def chunck_network(self, network):
        """ Create a list of Network chunks that can be pre-analyzed completely.
        Each chunk needs to be executed before the next can be analyzed and
        executed.

        The returned chunks are (at the moment) in the format of a tuple (start, nodes)
        which are both tuples. The tuple contain the nodes where to start execution (should
        ready if previous chunks are done) and all nodes of the chunk respectively.

        :param network: Network to split into chunks
        :return: tuple containing chunks
        """
        self._set_network_used_nodes(network)

        fastr.log.debug('START Pool: {}'.format(sorted([x.id for x in self.used_nodes])))
        candidates = deque()

        for source in network.sourcelist.values():
            if source in self.used_nodes:
                candidates.append(source)

        for constant in network.constantlist.values():
            if constant in self.used_nodes:
                candidates.append(constant)

        fastr.log.debug('START candidates: {}'.format([x.id for x in candidates]))

        chunks = []
        self.pool = set(self.used_nodes)
        while len(self.used_nodes) > 0:
            chunk_start = [x for x in candidates]
            new_chunk = []
            new_candidates = deque()
            new_pool = set()

            while len(candidates) > 0:
                node = candidates.popleft()
                fastr.log.debug('Considering Node {}'.format(node.id))

                if node in candidates:
                    fastr.log.error('NODE HAS ENTERED CANDIDATES TWICE! {}'.format(node))

                if node not in self.pool:
                    fastr.log.debug('Used nodes: {}\nPool: {}\nCandidates: {}\nNew pool: {}'
                                    '\nNew candidates: {}'.format([x.id for x in self.used_nodes],
                                                                  [x.id for x in self.pool],
                                                                  [x.id for x in candidates],
                                                                  [x.id for x in new_pool],
                                                                  [x.id for x in new_candidates]))
                    message = 'Encountered previously visited Node {}!'.format(node.id)
                    fastr.log.error(message)
                    raise exceptions.FastrStateError(message)

                # If this Node definitely cannot be executed, move to pool for next chunks
                if not self._node_is_candidate(node):
                    fastr.log.debug('Moving {} to next chunk (not a candidate)'.format(node.id))
                    new_candidates.append(node)
                    continue

                # If this Node is blocked by earlier nodes in the Chunk, move it to pool for next chunks
                if self._node_is_blocked(node):
                    fastr.log.debug('Moving {} to next chunk (is blocked)'.format(node.id))
                    new_candidates.append(node)
                    continue

                # Change node status and append to working chunk
                fastr.log.debug('Adding {} to chunk'.format(node.id))
                self.used_nodes.remove(node)
                self.node_status[node.id] = 'visited'
                new_chunk.append(node)

                fastr.log.debug('Processing listeners for {}'.format(node.id))
                # Recurse into following Nodes
                for listener in node.listeners:
                    lnode = listener.target.node
                    fastr.log.debug('Considering listener {}'.format(lnode.id))

                    if lnode not in self.used_nodes:
                        fastr.log.debug('Ignoring {}'.format(lnode.id))
                        continue

                    if not self._node_is_blocked(lnode):
                        if lnode not in candidates:
                            fastr.log.debug('Adding {} to candidates'
                                            ' (blocking {})'.format(lnode.id,
                                                                    self._node_is_blocked(lnode)))
                            candidates.append(lnode)
                        else:
                            fastr.log.debug('Listener {} already in candidates'.format(lnode.id))
                    else:
                        if lnode not in new_candidates:
                            fastr.log.debug('Queueing {} ({})'.format(lnode.id, self._node_is_blocked(lnode)))
                            if lnode in candidates:
                                candidates.remove(lnode)
                            new_candidates.append(lnode)

            # Add the newly created chunk to the list
            chunk_start = tuple(x for x in chunk_start if x in new_chunk)
            fastr.log.debug('Adding chunk {} with start {}'.format([x.id for x in new_chunk],
                                                                   [x.id for x in chunk_start]))

            if len(new_chunk) == 0 and candidates == new_candidates:
                raise exceptions.FastrStateError('Network chunker does not converge! It appears there is a bug!')

            chunks.append((chunk_start, tuple(new_chunk)))

            candidates = new_candidates
            self.processed_nodes.update(new_chunk)
            self.pool = self.pool - self.processed_nodes  # Remove processed nodes from pool
            fastr.log.debug('Start new chuck with candidates {} and pool {}'.format([x.id for x in candidates],
                                                                                    [x.id for x in self.pool]))

        # After chunking return a tuple of tuples, to avoid tempering later on
        return tuple(chunks)

    def _node_is_candidate(self, node):
        """
        Check if the Node is considered a candidate

        :param node: Node to check
        :return: flag indicating the Node is a candidate
        """
        for node in node.get_sourced_nodes():
            if node not in self.processed_nodes and node not in self.pool:
                return False

        return True

    @staticmethod
    def _node_is_analyzable(node):
        """
        Check if it is possible to analyze a Node

        :param node: Node to check
        :return: flag indicating the Node is analyzable
        """
        for snode in node.get_sourced_nodes():
            if snode.blocking:
                return False

        return True

    def _node_is_blocked(self, target):
        """
        Check if a Node is blocked

        :param target: Node to check
        :return:flag indicating the Node is blocked by a blocking Node
        """
        for node in target.get_sourced_nodes():
            if node in self.pool and (node.blocking or self._node_is_blocked(node)):
                fastr.log.debug('Node {} is the cause of the blocked node {} (pool: {})'.format(node.id,
                                                                                                target.id,
                                                                                                [x.id for x in self.pool]))
                return True

        return False

    def _set_network_used_nodes(self, network):
        """ Create a list of used Nodes from the network

        :param network: Network to analyze
        """
        self.used_nodes.clear()
        self.node_status.clear()
        for sink in network.sinklist.values():
            self.used_nodes.add(sink)
            if not self._node_is_analyzable(sink):
                self.node_status[sink.id] = 'blocked'
            else:
                self.node_status[sink.id] = 'unvisited'

            self._network_walker(sink)

    def _network_walker(self, node):
        """
        Recursive backwards search through the network to find all Nodes required for the Sinks.

        :param node: Node to start search from
        """
        # loop over inputs of the node
        for input_ in node.inputs.itervalues():
            # loop over the subinputs of every input
            for parentnode in input_.get_sourced_nodes():
                # if the node connected to the input is not the source node then add to execute list and walk further.
                #print('Adding parent {}'.format(parentnode.fullid))
                self.used_nodes.add(parentnode)
                if parentnode.blocking:
                    self.node_status[parentnode.id] = 'blocking'
                elif not self._node_is_analyzable(parentnode):
                    self.node_status[parentnode.id] = 'blocked'
                else:
                    self.node_status[parentnode.id] = 'unvisited'

                # Recurse down into the network
                self._network_walker(parentnode)




# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" FASTR is a top level package which includes all parts required to create
networks and edit networks.

.. class:: Network

    The class representing a Network, this is in fact a reference to
    :py:class:`fastr.core.network.Network`.

.. class:: Node

    The class representing a Node, this is in fact a reference to
    :py:class:`fastr.core.node.Node`.

.. class:: Link

    The class representing a Link, this is in fact a reference to
    :py:class:`fastr.core.link.Link`.

.. class:: SourceNode

    The class representing a data source, this is in fact a reference to
    :py:class:`fastr.core.node.SourceNode`.

.. class:: SinkNode

    The class representing a data sink, this is in fact a reference to
    :py:class:`fastr.core.node.SinkNode`.

.. class:: ConstantNode

    The class representing a constant data source, this is in fact a reference
    to :py:class:`fastr.core.node.ConstantNode`.

.. data:: toollist

    A :py:class:`fastr.core.toolmanager.ToolManager` containing all Tools known
    to the FASTR environment. The toollist can be accessed in a similar way to
    a dict. Indexing with a tool id will return the newest version of the Tool.
    If a specific version of the tool is required a tuple can be used as the
    index:

    .. code-block:: python

        >>> import fastr
        >>> fastr.toollist['testtool']
        <Tool: testtool version: 4.2>
        >>> fastr.toollist['testtool', '2.0']
        <Tool: testtool version: 2.0>


.. data:: typelist

    A :py:class:`fastr.core.datatypemanager.DataTypeManager` containing all
    Types known to the FASTR environment. This is usuable as a dict where
    the key is the datatype id and the value is the datatype itself.
"""

# In the top level module we want to add some variables which are constants
# but use a non-constant name (not caps)
# pylint: disable=invalid-name

from colorama import init
init()

from fastr.configmanager import Config

# Get version info
from fastr.version import version as __version__
from fastr import version

#: Configuration of the fastr system
config = Config()

# Reference the logger object from the config
log = config.log

__all__ = ['Network', 'Link', 'Node', 'ConstantNode', 'SourceNode', 'SinkNode'
           'typelist', 'datatypes', 'plugins', 'toollist', 'networklist',
           'ioplugins', 'vfs']

# Import all fastr components
from fastr.core.network import Network
from fastr.core.link import Link
from fastr.core.node import Node
from fastr.core.node import ConstantNode
from fastr.core.node import SourceNode
from fastr.core.node import SinkNode

from fastr.core.pluginmanager import LazyModule, PluginManager
from fastr.core.datatypemanager import typelist
from fastr.core.vfs import VirtualFileSystem

# Import datatypes and create a lazy loading module for that
import fastr.datatypes
datatypes = LazyModule("datatypes", parent=fastr.datatypes, plugin_manager=typelist)

plugin_manager = PluginManager()
import fastr.plugins
plugins = LazyModule("plugins", parent=fastr.plugins, plugin_manager=plugin_manager)

from fastr.core.interface import InterfacePluginManager
from fastr.core.ioplugin import IOPluginManager
from fastr.execution.executionpluginmanager import ExecutionPluginManager

# Load resource managers
from fastr.core.toolmanager import toollist
from fastr.core.networkmanager import networklist

# The following loads all ioplugins from the resources folder and registers the built-in vfs with the plugin list
ioplugins = IOPluginManager()
vfs = VirtualFileSystem()
execution_plugins = ExecutionPluginManager()

#: The currently active Network in fastr
current_network = None


def find(object_fullid):
    """Find a object by its fullid and return it.

    :param object_fullid: The full id of the object to retrieve
    :return: the desire object, or None if object not found
    """
    if current_network is None:
        log.info('No current network is set, cannot retrieve objects')
        return None

    try:
        return current_network[object_fullid]
    except (KeyError, IndexError):
        log.debug('Could not find "{}" in current network'.format(object_fullid))
        return None


# Register the val URL for Python version under 2.7.3
ioplugins.register_url_scheme('val')
interfaces = InterfacePluginManager()

log.debug('Finished with the FASTR environment set up')


# Warn if this is not a neatly installed package from the stable branch
if fastr.config.warn_develop:
    if version.not_default_branch or version.from_mercurial:
        fastr.log.warning('Not running in a production installation (branch "{}" from {})'.format(
            version.hg_branch,
            'source code' if version.from_mercurial else 'installed package'
        ))



import os
import shutil
import subprocess
import tempfile

FASTR_TOOLS_REPO = 'ssh://hg@bitbucket.org/bigr_erasmusmc/fastr-tools'


def main():
    try:
        # Clone the fastr-tools repository
        print('Creating temporary directory...')
        tmpdir = tempfile.mkdtemp()
        print('Created {}'.format(tmpdir))
        print('Cloning the fastr-tools repository...')
        hg_proc = subprocess.Popen(['hg', 'clone', FASTR_TOOLS_REPO], cwd=tmpdir)
        hg_proc.wait()
        print('Mercurial clone done...')
        print('Dirs in tmpdir: {}'.format(os.listdir(tmpdir)))
    
        # Copy the ~/.fastr/tools and ~/.fastr/types
        print('Copy tools...')
        tooldir = os.path.expanduser('~/.fastr/tools')
        if os.path.exists(tooldir):
            print('Removing old tools...')
            shutil.rmtree(tooldir)
        shutil.copytree(os.path.join(tmpdir, 'fastr-tools', 'image_analysis', 'tools'), tooldir)

        print('Copy types...')
        typesdir = os.path.expanduser('~/.fastr/datatypes')
        if os.path.exists(typesdir):
            print('Removing old types...')
            shutil.rmtree(typesdir)
        shutil.copytree(os.path.join(tmpdir, 'fastr-tools', 'image_analysis', 'datatypes'), typesdir)

    except OSError:
        print('Could not find/call mercurial correctly! Please make sure mercurial is installed!')
    finally:
        print('Cleaning up temporary directory')
        if 'tmpdir' in locals():
            shutil.rmtree(tmpdir)

if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import urlparse

import prov
from prov.model import ProvDocument

import fastr
from fastr import exceptions
from fastr.data import url
from fastr.datatypes import Deferred
from fastr.utils import iohelpers


class Provenance(object):
    """
    The Provenance object keeps track of everything that happens to a data object.
    """

    def __init__(self, host=None):
        if host is not None:
            self.host = host.rstrip('/')
        else:
            self.host = fastr.config.web_url()
        self.namespaces = {}
        self.document = ProvDocument()

        # Define default namespaces
        self.fastr = self._add_namespace('fastr')
        self.tool = self._add_namespace('tool')
        self.node = self._add_namespace('node')
        self.job = self._add_namespace('job')
        self.data = self._add_namespace('data')
        self.worker = self._add_namespace('worker')
        self.network = self._add_namespace('network')
        self.fastr_info = self._add_namespace('fastrinfo')

        # Add fastr agent
        self.fastr_agent = self.agent(self.fastr[fastr.version.version])

    def _add_namespace(self, name, parent=None, url=None):
        if parent is None:
            host = self.host
        else:
            if parent in self.namespaces.keys():
                host = self.namespaces[parent].uri
            else:
                return None

        if url is None:
            self.namespaces[name] = self.document.add_namespace(name, "{}/{}/".format(host, name))
        else:
            self.namespaces[name] = self.document.add_namespace(name, "{}/{}".format(host, url))
        return self.namespaces[name]

    def agent(self, identifier, other_attributes=None):
        return self.document.agent(identifier, other_attributes)

    def activity(self, identifier, start_time=None, end_time=None, other_attributes=None):
        return self.document.activity(identifier, start_time, end_time, other_attributes)

    def entity(self, identifier, other_attributes=None):
        return self.document.entity(identifier, other_attributes)

    def init_provenance(self, job):
        """
        Create initial provenance document
        """
        self.job_activity = self.activity(self.job[job.id])
        self.tool_agent = self.agent(
            self.tool["{}/{}".format(job.tool_name, job.tool_version)],
            {
                'fastrinfo:tool_dump': fastr.toollist[job.tool_name, job.tool_version].dumps()
            }
        )
        self.node_agent = self.agent(self.node[job.node_id])

        #FIXME: Find a better way to collect the current network id
        self.network_agent = self.agent(self.network["{}/{}".format(job.network_id, job.network_version)])

        self.document.actedOnBehalfOf(self.network_agent, self.fastr_agent)
        self.document.actedOnBehalfOf(self.node_agent, self.tool_agent)
        self.document.actedOnBehalfOf(self.node_agent, self.network_agent)
        self.document.wasAssociatedWith(self.job_activity, self.node_agent)

    def collect_provenance(self, job, advanced_flow=False):
        """
        Collect the provenance for this job
        """
        tool = job.tool
        for input_argument_key, input_argument_value in job.input_arguments.items():
            input_description = tool.inputs[input_argument_key]

            if input_description.hidden:
                # Skip hidden inputs (they are used for the system to feed
                # additional data to sources/sinks etc) not interesting
                # for provenance
                continue

            if not advanced_flow:
                self.collect_input_argument_provenance(input_argument_value)
            else:
                for sample in input_argument_value:
                    self.collect_input_argument_provenance(sample)

        self.document._records = list(set(self.document._records))

    def collect_input_argument_provenance(self, input_argument):
        for cardinality, value in enumerate(input_argument.data.iterelements()):
            parent_provenance, parent_job = self.get_parent_provenance(value)
            value_url = self.data_uri(value, parent_job)

            if isinstance(parent_provenance, ProvDocument):
                self.document.update(parent_provenance)
                data_entity = self.entity(self.data[value_url], {
                    'fastrinfo:value': value.value
                })
                self.document.wasGeneratedBy(data_entity, self.activity(self.job[parent_job.id]))
            else:
                data_entity = self.entity(self.data[value_url])
            self.document.used(self.job_activity, data_entity)

    @staticmethod
    def data_uri(value, job):
        fastr.log.info('DATA URI for: {}, job: {}'.format(value.raw_value, job))
        if isinstance(value.raw_value, str) and value.raw_value.startswith('val://'):
            val_uri = urlparse.urlparse(value.raw_value)
            val_query = urlparse.parse_qs(val_uri.query)
            sample_id = val_query['sampleid'][0] if 'sampleid' in val_query else job.sample_id

            return 'fastr://data/job/{job_id}/output/{outputid}/sample/{sample}/cardinality/{cardinality}'.format(
                job_id=job.id,
                outputid=val_query['outputname'][0],
                sample=sample_id,
                cardinality=val_query['nr'][0]
            )
        else:
            if url.isurl(value.raw_value):
                return value.raw_value
            else:
                return 'fastr://data/constant/{}'.format(value.raw_value)

    @staticmethod
    def get_parent_provenance(value):
        """
        Find the provenance of the parent job

        :param str value: url for the value for which to find the job
        :return: the provenance of the job that created the value
        :raises FastrKeyError: if the deferred is not available (yet)
        :raises FastrValueError: if the value is not a valid deferred url
        """
        if not isinstance(value, Deferred):
            return None, None

        parsed_url = urlparse.urlparse(value.raw_value)

        if parsed_url.scheme != 'val':
            raise exceptions.FastrValueError('Cannot lookup value {}, wrong url scheme'.format(value.raw_value))

        # First load job to find location of the prov file
        datafile = os.path.join(fastr.config.mounts[parsed_url.netloc],
                                os.path.normpath(parsed_url.path[1:]))

        # Open Job file
        data = iohelpers.load_gpickle(datafile)

        # Get provenance file
        prov.read(data.provfile, 'json')

        return prov, data

    def serialize(self, filename, format):
        self.document.serialize(destination=filename, format=format)



import fastr

IS_TEST = False

def create_network():
    network = fastr.Network('auto_prefix_test')

    const = network.create_constant(fastr.datatypes.Int, [1,2,3], id_='const')
    source = network.create_source(fastr.datatypes.Int, id_='source')

    node_m_p = network.create_node('AutoPrefixTest', id_='m_p')
    node_a_p = network.create_node('AutoPrefixTest', id_='a_p')
    node_m_n = network.create_node('AutoPrefixNegateTest', id_='m_n')
    node_a_n = network.create_node('AutoPrefixNegateTest', id_='a_n')

    sink_m_p = network.create_sink(fastr.datatypes.Int, id_='sink_m_p')
    sink_a_p = network.create_sink(fastr.datatypes.Int, id_='sink_a_p')
    sink_m_n = network.create_sink(fastr.datatypes.Int, id_='sink_m_n')
    sink_a_n = network.create_sink(fastr.datatypes.Int, id_='sink_a_n')

    for node in [node_m_p, node_a_p, node_m_n, node_a_n]:
        node.inputs['left_hand'] = source.output
        node.inputs['right_hand'] = const.output
        

    sink_m_p.input = node_m_p.outputs['multiplied']
    sink_m_n.input = node_m_n.outputs['multiplied']
    sink_a_n.input = node_a_n.outputs['added']
    sink_a_p.input = node_a_p.outputs['added']

    return network


def main():
    network = create_network()
    network.draw_network()
    network.dumpf('auto_prefix_test.json')

    network.execute(sourcedata={'source': [5]},
                    sinkdata={
                        'sink_a_n': 'vfs://tmp/results/auto_prefix/{node}_{sample_id}',
                        'sink_a_p': 'vfs://tmp/results/auto_prefix/{node}_{sample_id}',
                        'sink_m_p': 'vfs://tmp/results/auto_prefix/{node}_{sample_id}',
                        'sink_m_n': 'vfs://tmp/results/auto_prefix/{node}_{sample_id}'
                        }
                    )

if __name__ == '__main__':
    main()



import argparse


def main():
    parser = argparse.ArgumentParser(description='Short sample app that create a range (in the cardinality)')
    parser.add_argument('--input', type=int)

    args = parser.parse_args()

    print('arguments: {}'.format(args))
    print('input: {}'.format(args.input))

    print('RESULT_1={}'.format(range(args.input)))


if __name__ == "__main__":
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Network module containing Network facilitators and analysers.
"""
import datetime
import functools
import json
import os
import re
import subprocess
import sys
import time
import urlparse
import uuid
import pprint
import traceback
import shutil

from collections import OrderedDict
from tempfile import mkdtemp
from threading import Event, Lock, RLock

import requests

import fastr
from fastr.core.link import Link
from fastr.core.node import Node, ConstantNode, SourceNode, SinkNode, MacroNode
from fastr.core.inputoutput import Output
from fastr.core.samples import SampleItem
from fastr.core.serializable import Serializable
from fastr.core.version import Version
from fastr.data import url
import fastr.exceptions as exceptions
from fastr.execution.networkchunker import DefaultNetworkChunker
from fastr.execution.networkanalyzer import DefaultNetworkAnalyzer
from fastr.execution.job import JobState
from fastr.utils import iohelpers, pim_publisher


class Network(Serializable):
    """
    The Network class represents a workflow. This includes all Nodes (including ConstantNodes, SourceNodes and Sinks) and
    Links.
    """

    __dataschemafile__ = 'Network.schema.json'

    NETWORK_DUMP_FILE_NAME = '__fastr_network__.json'
    SOURCE_DUMP_FILE_NAME = '__source_data__.pickle.gz'
    SINK_DUMP_FILE_NAME = '__sink_data__.json'

    def __init__(self, id_='unnamed_network', version=None):
        """
        Create a new, empty Network

        :param str name: name of the Network
        :return: newly created Network
        :raises OSError: if the tmp mount in the config is not a writable directory
        """

        regex = r'^\w[\w\d_]*$'
        if re.match(regex, id_) is None:
            raise exceptions.FastrValueError('An id in Fastr should follow'
                                             ' the following pattern {}'
                                             ' (found {})'.format(regex, id_))

        if version is None:
            version = '0.0'

        #: The namespace this network lives in, this will be set by the NetworkManager on load
        self.namespace = None

        self._id = id_
        self.version = Version(version)
        self.description = ''
        self.toolnodelist = {}
        self.nodelist = {}
        self.sinklist = {}
        self.constantlist = {}
        self.sourcelist = {}
        self.macrolist = {}
        self.linklist = {}
        self.preferred_types = []
        self.stepids = {}
        self.nodegroups = {}
        self.link_number = 0
        self.node_number = 0
        self.callback_edit_lock = Lock()
        self.executing = False
        self.execution_lock = Event()
        self.execution_lock.set()
        self.edit_lock = RLock()

        self.job_finished_callback = None
        self.job_status_callback = None

        self.filename = None
        self.sink_results = {}

        # Info about the execution run (should be moved to NetworkRun in the future)
        self.timestamp = None
        self.uuid = None
        self.tmpdir = None
        self.tmpurl = None

        # Check if temp dir exists, if not try to create it
        if not os.path.exists(fastr.config.mounts['tmp']):
            fastr.log.info("fast temporary directory does not exists, creating it...")
            try:
                os.mkdir(fastr.config.mounts['tmp'])
            except OSError:
                message = "Could not create fastr temporary directory ({})".format(fastr.config.mounts['tmp'])
                fastr.log.critical(message)
                raise exceptions.FastrOSError(message)

        fastr.log.info('Changing fastr.current network to "{}"'.format(self.id))
        fastr.current_network = self

    def __repr__(self):
        return '<Network {} (v{})>'.format(self.id, self.version)

    def __eq__(self, other):
        """
        Compare two Networks and see if they are equal.

        :param other:
        :type other: :py:class:`Network <fastr.core.network.Network>`
        :return: flag indicating that the Networks are the same
        :rtype: bool
        """
        if not isinstance(other, Network):
            return NotImplemented

        dict_self = {k: v for k, v in self.__dict__.items()}
        del dict_self['callback_edit_lock']
        del dict_self['execution_lock']
        del dict_self['edit_lock']
        del dict_self['job_finished_callback']
        del dict_self['job_status_callback']
        del dict_self['tmpdir']
        del dict_self['tmpurl']
        del dict_self['nodelist']
        del dict_self['uuid']
        del dict_self['timestamp']
        del dict_self['filename']
        del dict_self['sink_results']

        dict_other = {k: v for k, v in other.__dict__.items()}
        del dict_other['callback_edit_lock']
        del dict_other['execution_lock']
        del dict_other['edit_lock']
        del dict_other['job_finished_callback']
        del dict_other['job_status_callback']
        del dict_other['tmpdir']
        del dict_other['tmpurl']
        del dict_other['nodelist']
        del dict_other['uuid']
        del dict_other['timestamp']
        del dict_other['filename']
        del dict_other['sink_results']

        return dict_self == dict_other

    def __ne__(self, other):
        """
        Tests for non-equality, this is the negated version __eq__
        """
        return not (self.__eq__(other))

    # Retrieve a Node/Link/Input/Output in the network based on the fullid
    def __getitem__(self, item):
        """
        Get an item by its fullid. The fullid can point to a link, node, input, output or even subinput/suboutput.

        :param str,unicode item: fullid of the item to retrieve
        :return: the requested item
        """
        if not isinstance(item, (str, unicode)):
            raise exceptions.FastrTypeError('Key should be a fullid string, found a {}'.format(type(item).__name__))

        parsed = urlparse.urlparse(item)
        if parsed.scheme != 'fastr':
            raise exceptions.FastrValueError('Item should be an URL with the fastr:// scheme (Found {} in {})'.format(parsed.scheme, item))

        path = parsed.path.split('/')[1:]

        if len(path) < 2 or path[0] != 'networks' or path[1] != self.id:
            raise exceptions.FastrValueError('URL {} does not point to anything in this network, {}'.format(item, path))

        value = self

        for part in path[2:]:
            if hasattr(value, '__getitem__'):
                try:
                    if isinstance(value, (list, tuple, Output)):
                        value = value[int(part)]
                    else:
                        value = value[part]
                except (KeyError, IndexError, TypeError, ValueError):
                    pass
                else:
                    continue

            if hasattr(value, part):
                value = getattr(value, part)
            else:
                raise exceptions.FastrLookupError('Could not find {} in {}'.format(part, value))

        return value

    def __getstate__(self):
        """
        Retrieve the state of the Network

        :return: the state of the object
        :rtype dict:
        """
        state = {
            'id': self.id,
            'version': str(self.version),
            'filename': self.filename,
            'description': self.description,
            'link_number': self.link_number,
            'node_number': self.node_number,
            'nodelist': [x.__getstate__() for x in self.nodelist.values()],
            'linklist': [x.__getstate__() for x in self.linklist.values()],
            'preferred_types': [x.id for x in self.preferred_types],
            'stepids': {k: [x.id for x in v] for k, v in self.stepids.items()},
            'nodegroups': self.nodegroups,
            'namespace': self.namespace
        }

        return state

    def __setstate__(self, state):
        """
        Set the state of the Network by the given state. This completely
        overwrites the old state!

        :param dict state: The state to populate the object with
        :return: None
        """
        # Initialize empty to avoid errors further on
        self._id = state['id']
        self.version = Version(state['version'])
        self.nodelist = {}
        self.linklist = {}
        self.macrolist = {}
        self.sourcelist = {}
        self.constantlist = {}
        self.sinklist = {}
        self.toolnodelist = {}
        self.preferred_types = []
        self.stepids = {}
        self.callback_edit_lock = Lock()
        self.execution_lock = Event()
        self.execution_lock.set()
        self.executing = False
        self.edit_lock = RLock()

        # There are run dependent, and can safely be lost during serialization
        self.job_finished_callback = None
        self.job_status_callback = None

        self.tmpdir = None
        self.tmpurl = None
        self.description = state['description']
        self.nodegroups = state['nodegroups']
        self.sink_results = {}

        # Info about the execution run (should be moved to NetworkRun in the future)
        self.timestamp = None
        self.uuid = None

        # Make proper version
        state['version'] = Version(state['version'])

        # Make sure the locks exist
        if not hasattr(self, 'callback_edit_lock'):
            self.callback_edit_lock = Lock()
        if not hasattr(self, 'execution_lock'):
            self.execution_lock = Event()
            self.execution_lock.set()
        if not hasattr(self, 'edit_lock'):
            self.edit_lock = RLock()

        # Set self as current network for links etc to be created properly
        fastr.current_network = self

        # Set ID, we need this for messages later on
        self._id = state['id']
        del state['id']

        # These should not be shared between Networks
        state.pop('callback_edit_lock', None)
        state.pop('execution_lock', None)
        state.pop('edit_lock', None)

        # Recreate nodes
        if 'nodelist' in state:
            for node_state in state['nodelist']:
                # Get the node class
                node_class = node_state.get('class', 'Node')
                node_class = getattr(fastr.core.node, node_class)

                node = node_class.createobj(node_state, self)
                fastr.log.debug('Adding node: {}'.format(node))
                self.add_node(node)
            del state['nodelist']

        # Add preferred types
        state['preferred_types'] = [fastr.typelist[x] for x in state['preferred_types']]

        # Insert empty link_list
        statelinklist = state['linklist']
        state['linklist'] = {}

        # Update the objects dict
        self.__dict__.update(state)

        # Create the link list, make sure all Nodes are in place first
        for link in statelinklist:
            self.linklist[link['id']] = Link.createobj(link, self)

        # Make the stepids reference the Node instead of using ids
        if self.stepids is None:
            self.stepids = {}
        self.stepids = {k: [self.nodelist[x] for x in v] for k, v in self.stepids.items()}

        self.node_number = state['node_number']
        self.link_number = state['link_number']

    @property
    def id(self):
        """
        The id of the Network. This is a read only property.
        """
        return self._id

    @property
    def fullid(self):
        """
        The fullid of the Network
        """
        return 'fastr:///networks/{}'.format(self.id)

    @property
    def run_id(self):
        """
        The run id of the network
        """
        return '{}_{}'.format(self.id, self.timestamp.strftime('%Y_%m_%d_%H_%M_%S_%f'))

    def add_node(self, node):
        """
        Add a Node to the Network. Make sure the node is in the node list and
        the node parent is set to this Network

        :param node: node to add
        :type node: :py:class:`Node <fastr.core.node.Node>`
        :raises FastrTypeError: if node is incorrectly typed
        """
        # Make sure we are not in the executing state
        if not self.execution_lock.wait(timeout=0):
            message = 'this network is being executed, cannot edit it during the execution'
            fastr.log.warning(message)
            return

        # Acquire the edit lock
        self.edit_lock.acquire()

        if node.id not in self.nodelist and isinstance(node, Node):
            self.nodelist[node.id] = node

            # Automatically sort Nodes in the right dict
            if isinstance(node, ConstantNode):
                self.constantlist[node.id] = node
            elif isinstance(node, SourceNode):
                self.sourcelist[node.id] = node
            elif isinstance(node, SinkNode):
                self.sinklist[node.id] = node
            elif isinstance(node, MacroNode):
                self.macrolist[node.id] = node
            elif isinstance(node, Node):
                self.toolnodelist[node.id] = node
            else:
                raise exceptions.FastrTypeError('Unknown Node type encountered! (type {})'.format(type(node).__name__))

            node.parent = self

        # Release the edit lock
        self.edit_lock.release()

    def add_link(self, link):
        """
        Add a Link to the Network. Make sure the link is in the link list and
        the link parent is set to this Network

        :param link: link to add
        :type link: :py:class:`Link <fastr.core.link.Link>`
        :raises FastrTypeError: if link is incorrectly typed
        :raises FastrNetworkMismatchError: if the link already belongs to another Network
        """

        if not isinstance(link, Link):
            raise exceptions.FastrTypeError('Link argument is not of Link class! (type {})'.format(type(link).__name__))

        # Make sure we are not in the executing state
        if not self.execution_lock.wait(timeout=0):
            message = 'this network is being executed, cannot edit it during the execution'
            fastr.log.warning(message)
            return

        # Acquire the edit lock
        self.edit_lock.acquire()

        if link.id not in self.linklist:
            if link.parent is None:
                # Make sure parent and network have mutual understanding of the arrangement
                link.parent = self
            elif link.parent is not self:
                self.edit_lock.release()  # Release lock  before raising exception
                raise exceptions.FastrNetworkMismatchError('Cannot add a Link that already belongs to another Network!')

            self.linklist[link.id] = link

        # Release the edit lock
        self.edit_lock.release()

    def remove(self, value):
        """
        Remove an item from the Network.

        :param value: the item to remove
        :type value: :py:class:`Node <fastr.core.node.Node>` or
                      :py:class:`Link <fastr.core.link.Link>`
        """
        # Make sure we are not in the executing state
        if not self.execution_lock.wait(timeout=0):
            message = 'this network is being executed, cannot edit it during the execution'
            fastr.log.warning(message)
            return

        # Acquire the edit lock
        self.edit_lock.acquire()

        if isinstance(value, Link):
            self.linklist.pop(value.id)

        if isinstance(value, Node):
            self.nodelist.pop(value.id)

        # Release the edit lock
        self.edit_lock.release()

    def add_stepid(self, stepid, node):
        """
        Add a Node to a specific step id

        :param str stepid: the stepid that the node will be added to
        :param node: the node to add to the stepid
        :type node: :py:class:`Node <fastr.core.node.Node>`
        """
        # Make sure we are not in the executing state
        if not self.execution_lock.wait(timeout=0):
            message = 'this network is being executed, cannot edit it during the execution'
            fastr.log.warning(message)
            return

        # Acquire the edit lock
        self.edit_lock.acquire()

        if stepid is not None:
            if stepid in self.stepids:
                self.stepids[stepid] += [node]
            else:
                self.stepids[stepid] = [node]

        # Release the edit lock
        self.edit_lock.release()

    def check_id(self, id_):
        """
        Check if an id for an object is valid and unused in the Network. The
        method will always returns True if it does not raise an exception.

        :param str id_: the id to check
        :return: True
        :raises FastrValueError: if the id is not correctly formatted
        :raises FastrValueError: if the id is already in use
        """

        regex = r'^\w[\w\d_]*$'
        if re.match(regex, id_) is None:
            raise exceptions.FastrValueError('An id in Fastr should follow'
                                             ' the following pattern {}'
                                             ' (found {})'.format(regex, id_))

        if id_ in self.nodelist or id_ in self.linklist:
            raise exceptions.FastrValueError('The id {} is already in use in {}!'.format(id_, self.id))

        return True

    def create_node(self, tool, id_=None, stepid=None, cores=None, memory=None, walltime=None, nodegroup=None):
        """
        Create a Node in this Network. The Node will be automatically added to
        the Network.

        :param tool: The Tool to base the Node on
        :type tool: :py:class:`Tool <fastr.core.tool.Tool>`
        :param str id_: The id of the node to be created
        :param str stepid: The stepid to add the created node to
        :param str nodegroup: The group the node belongs to, this can be
                              important for FlowNodes and such, as they
                              will have matching dimension names.
        :return: the newly created node
        :rtype: :py:class:`Node <fastr.core.node.Node>`
        """
        # Make sure we are not in the executing state
        if not self.execution_lock.wait(timeout=0):
            message = 'this network is being executed, cannot edit it during the execution'
            fastr.log.warning(message)
            return

        # Acquire the edit lock
        self.edit_lock.acquire()

        # Create a node in the network
        if isinstance(tool, (str, tuple)):
            tool = fastr.toollist[tool]

        try:
            NodeType = getattr(fastr.core.node, tool.node_class)
        except AttributeError:
            raise exceptions.FastrValueError('The indicated node class {} cannot be found for Tool {}/{}'.format(tool.node_class, tool.id, tool.version))

        node = NodeType(tool, id_, self, cores=cores, memory=memory, walltime=walltime)
        self.add_node(node)
        self.add_stepid(stepid, node)
        if nodegroup is None:
            nodegroup = node.id

        if nodegroup not in self.nodegroups:
            self.nodegroups[nodegroup] = []
        self.nodegroups[nodegroup].append(node.id)

        # Release the edit lock
        self.edit_lock.release()

        return node

    def create_macro(self, network, id_=None):
        if not self.execution_lock.wait(timeout=0):
            message = 'this network is being executed, cannot edit it during the execution'
            fastr.log.warning(message)
            return
        self.edit_lock.acquire()
        node = MacroNode(network, id_, self)

        self.add_node(node)
        self.edit_lock.release()
        return node

    def create_constant(self, datatype, data, id_=None, stepid=None, nodegroup=None, sourcegroup=None):
        """
        Create a ConstantNode in this Network. The Node will be automatically added to
        the Network.

        :param datatype: The DataType of the constant node
        :type datatype: :py:class:`BaseDataType <fastr.core.datatypemanager.BaseDataType>`
        :param data: The data to hold in the constant node
        :type data: datatype or list of datatype
        :param str id_: The id of the constant node to be created
        :param str stepid: The stepid to add the created constant node to
        :param str nodegroup: The group the node belongs to, this can be
                              important for FlowNodes and such, as they
                              will have matching dimension names.
        :return: the newly created constant node
        :rtype: :py:class:`ConstantNode <fastr.core.node.ConstantNode>`
        """
        # Make sure we are not in the executing state
        if not self.execution_lock.wait(timeout=0):
            message = 'this network is being executed, cannot edit it during the execution'
            fastr.log.warning(message)
            return

        # Acquire the edit lock
        self.edit_lock.acquire()

        if not isinstance(data, (list, dict, OrderedDict)):
            data = [data]
        const_node = ConstantNode(datatype, data, id_)
        self.add_node(const_node)
        self.add_stepid(stepid, const_node)
        if nodegroup is None:
            if sourcegroup is None:
                nodegroup = const_node.id
            else:
                fastr.log.warning('[DEPRECATED] The sourcegroup kwarg of the'
                                  ' Network.create_source is deprecated and'
                                  ' replaced by the nodegroup kwarg. Please'
                                  ' use that kwarg instead, it will have the'
                                  ' same function')
                nodegroup = sourcegroup

        if nodegroup not in self.nodegroups:
            self.nodegroups[nodegroup] = []
        self.nodegroups[nodegroup].append(const_node.id)

        # Release the edit lock
        self.edit_lock.release()

        return const_node

    def create_link(self, source, target, id_=None, collapse=None, expand=None):
        """
        Create a link between two Nodes and add it to the current Network.

        :param source: the output that is the source of the link
        :type source: :py:class:`BaseOutput <fastr.core.inputoutput.BaseOutput>`
        :param target: the input that is the target of the link
        :type target: :py:class:`BaseInput <fastr.core.inputoutput.BaseInput>`
        :param str id_: the id of the link
        :return: the created link
        :type: :py:class:`Link <fastr.core.link.Link>`
        """
        # Make sure we are not in the executing state
        if not self.execution_lock.wait(timeout=0):
            message = 'this network is being executed, cannot edit it during the execution'
            fastr.log.warning(message)
            return

        # Acquire the edit lock
        self.edit_lock.acquire()

        link = Link(source, target, id_=id_, parent=self, collapse=collapse, expand=expand)
        self.add_link(link)

        # Release the edit lock
        self.edit_lock.release()

        return link

    def create_source(self, datatype, id_=None, stepid=None, nodegroup=None, sourcegroup=None):
        """
        Create a SourceNode in this Network. The Node will be automatically added to
        the Network.

        :param datatype: The DataType of the source source_node
        :type datatype: :py:class:`BaseDataType <fastr.core.datatypemanager.BaseDataType>`
        :param str id_: The id of the source source_node to be created
        :param str stepid: The stepid to add the created source source_node to
        :param str nodegroup: The group the node belongs to, this can be
                              important for FlowNodes and such, as they
                              will have matching dimension names.
        :param str sourcegroup: *DEPRECATED!* The nodegroup this SourceNode will be added to
        :return: the newly created source source_node
        :rtype: :py:class:`SourceNode <fastr.core.source_node.SourceNode>`
        """
        # Make sure we are not in the executing state
        if not self.execution_lock.wait(timeout=0):
            message = 'this network is being executed, cannot edit it during the execution'
            fastr.log.warning(message)
            return

        # Acquire the edit lock
        self.edit_lock.acquire()

        # Set a source for the network.
        source_node = SourceNode(datatype=datatype, id_=id_)
        self.add_node(source_node)
        self.add_stepid(stepid, source_node)

        if nodegroup is None:
            if sourcegroup is None:
                nodegroup = source_node.id
            else:
                fastr.log.warning('[DEPRECATED] The sourcegroup kwarg of the'
                                  ' Network.create_source is deprecated and'
                                  ' replaced by the nodegroup kwarg. Please'
                                  ' use that kwarg instead, it will have the'
                                  ' same function')
                nodegroup = sourcegroup

        if nodegroup not in self.nodegroups:
            self.nodegroups[nodegroup] = []
        self.nodegroups[nodegroup].append(source_node.id)

        # Release the edit lock
        self.edit_lock.release()

        return source_node

    def create_sink(self, datatype, id_=None, stepid=None):
        """
        Create a SinkNode in this Network. The Node will be automatically added to
        the Network.

        :param datatype: The DataType of the sink node
        :type datatype: :py:class:`BaseDataType <fastr.core.datatypemanager.BaseDataType>`
        :param str id_: The id of the sink node to be created
        :param str stepid: The stepid to add the created sink node to
        :return: the newly created sink node
        :rtype: :py:class:`SinkNode <fastr.core.node.SinkNode>`
        """
        # Make sure we are not in the executing state
        if not self.execution_lock.wait(timeout=0):
            message = 'this network is being executed, cannot edit it during the execution'
            fastr.log.warning(message)
            return

        # Acquire the edit lock
        self.edit_lock.acquire()

        # Set a sink for the network
        node = SinkNode(datatype=datatype, id_=id_)
        self.add_node(node)
        self.add_stepid(stepid, node)

        # Release the edit lock
        self.edit_lock.release()

        return node

    def create_reference(self, source_data, output_directory):
        # Create temporary output directory
        if not os.path.exists(output_directory):
            os.makedirs(output_directory)

        output_url = fastr.vfs.path_to_url(output_directory, scheme='ref')
        fastr.log.info('Created temp result directory {}'.format(output_directory))
        network_temp_dir = os.path.join(output_directory, '__fastr_run_tmp__')

        # Save source data to reference dir
        iohelpers.save_gpickle(os.path.join(output_directory, self.SOURCE_DUMP_FILE_NAME), source_data)

        # Save a dump of the network
        with open(os.path.join(output_directory, self.NETWORK_DUMP_FILE_NAME), 'w') as fh_out:
            self.dump(fh_out, method='json', indent=2)

        # Set the output sink data to temporary directory
        sink_data = {}
        for sink in self.sinklist.values():
            sink_data[sink.id] = '{}/{}/{{sample_id}}/{{cardinality}}/result.json'.format(
                output_url,
                sink.id
            )
        fastr.log.info('Set sink data to: {}'.format(sink_data))

        # Execute the network
        self.execute(sourcedata=source_data, sinkdata=sink_data, tmpdir=network_temp_dir)

    def test(self, reference_data_dir, source_data=None):
        """
        Execute the network with the source data specified and test the results
        against the refence data. This effectively tests the network execution.

        :param str reference_data_dir: The path or vfs url of reference data to compare with
        :param dict source_data: The source data to use
        """
        if not isinstance(reference_data_dir, basestring):
            raise exceptions.FastrTypeError('reference_data_dir should be a string!')

        if reference_data_dir.startswith('vfs://'):
            reference_data_dir = fastr.vfs.url_to_path(reference_data_dir)

        if not os.path.isdir(reference_data_dir):
            raise exceptions.FastrTypeError('The reference_data_dir should be pointing to an existing directory!')

        if source_data is None:
            source_data = iohelpers.load_gpickle(os.path.join(reference_data_dir, self.SOURCE_DUMP_FILE_NAME))

        temp_results_dir = None
        validation_result = True
        try:
            # Create temporary output directory
            temp_results_dir = os.path.normpath(mkdtemp(prefix='fastr_network_test_'.format(self.id), dir=fastr.config.mounts['tmp']))
            self.create_reference(source_data=source_data, output_directory=temp_results_dir)

            # Check all sinks results
            for sink in self.sinklist.values():
                sink_output_dir = os.path.join(temp_results_dir, sink.id)
                sink_reference_dir = os.path.join(reference_data_dir, sink.id)

                output_samples = sorted(os.listdir(sink_output_dir))
                reference_samples = sorted(os.listdir(sink_reference_dir))

                if output_samples != reference_samples:
                    fastr.log.info('Different samples found in sink "{}"'.format(sink.id))
                    fastr.log.info('Output samples: {}'.format(output_samples))
                    fastr.log.info('Reference samples: {}'.format(reference_samples))
                    validation_result = False
                    continue

                for sample in output_samples:
                    sample_output_dir = os.path.join(sink_output_dir, sample)
                    sample_reference_dir = os.path.join(sink_reference_dir, sample)

                    output_values = sorted(os.listdir(sample_output_dir))
                    reference_values = sorted(os.listdir(sample_reference_dir))

                    if output_values != reference_values:
                        fastr.log.info('Difference number of cardinality entries for {}/{}'.format(sink.id, sample))
                        fastr.log.info('Output cardinality entries: {}'.format(output_values))
                        fastr.log.info('Reference cardinality entries: {}'.format(reference_values))
                        validation_result = False
                        continue

                    for value in output_values:
                        with open(os.path.join(sample_output_dir, value, 'result.json')) as fh_in:
                            output_data = json.load(fh_in)
                        with open(os.path.join(sample_reference_dir, value, 'result.json')) as fh_in:
                            reference_data = json.load(fh_in)

                        output_item = fastr.typelist[output_data['datatype']](output_data['value'])
                        reference_item = fastr.typelist[reference_data['datatype']](reference_data['value'])

                        if output_item != reference_item:
                            fastr.log.info('Value for {}/{}/{} was not equal! (found "{}", expected "{}")'.format(
                                sink.id,
                                sample,
                                value,
                                output_item,
                                reference_item)
                            )

                            fastr.log.info('Output: [{}] {!r}'.format(type(output_item.value).__name__,
                                                                      output_item.value))
                            fastr.log.info('Reference: [{}] {!r}'.format(type(reference_item.value).__name__,
                                                                         reference_item.value))

                            validation_result = False

            if validation_result:
                fastr.log.info('Run and reference were equal! Test passed!')
            return validation_result
        finally:
            # Clean up
            fastr.log.info('Removing temp result directory {}'.format(temp_results_dir))
            if temp_results_dir is not None and os.path.isdir(temp_results_dir):
                shutil.rmtree(temp_results_dir, ignore_errors=True)

    def execute(self, sourcedata, sinkdata, execution_plugin=None, tmpdir=None, cluster_queue=None):
        """
        Execute the Network with the given data. This will analyze the Network,
        create jobs and send them to the execution backend of the system.

        :param dict sourcedata: dictionary containing all data for the sources
        :param dict sinkdata: dictionary containing directives for the sinks
        :param str execution_plugin: the execution plugin to use (None will use the config value)
        :raises FastrKeyError: if a source has not corresponding key in sourcedata
        :raises FastrKeyError: if a sink has not corresponding key in sinkdata
        """
        result = False

        self.timestamp = datetime.datetime.now()
        self.uuid = uuid.uuid1()

        with self.edit_lock:
            fastr.log.debug('Acquiring execution lock...')
            self.execution_lock.wait()
            self.execution_lock.clear()
            self.executing = True

            fastr.log.info("####################################")
            fastr.log.info("#     network execution STARTED    #")
            fastr.log.info("####################################")

            # Display the main file the Network is created from, if known it
            # is the network.filename, otherwise try to estimate the entry
            # point for the script used
            entry_file = self.filename

            if entry_file is None:
                for x in sys.argv[:2]:
                    if os.path.isfile(x):
                        entry_file = x
                        break

            # If there is no entry_file or the first argument is ipython or
            # bpython we assume that is not a network but via prompt
            if entry_file is not None and not entry_file.endswith((os.path.sep + "ipython",
                                                                   os.path.sep + "bpython")):
                entry_file = os.path.abspath(entry_file)
                entry_file_date = time.ctime(os.path.getmtime(entry_file))
                message = 'Running network via {} (last modified {})'.format(entry_file, entry_file_date)
            else:
                if entry_file is not None:
                    entry_file = os.path.basename(entry_file)
                message = 'Running network via {} session'.format(entry_file or 'python')

            fastr.log.info(message)
            fastr.log.info('FASTR loaded from {}'.format(fastr.__file__))

            if tmpdir is None:
                tmpdir = os.path.normpath(mkdtemp(prefix='fastr_{}_'.format(self.id), dir=fastr.config.mounts['tmp']))
            else:
                if url.isurl(tmpdir):
                    if not tmpdir.startswith('vfs://'):
                        raise exceptions.FastrValueError('The tmpdir keyword argument should be a path or vfs:// url!')
                    tmpdir = fastr.vfs.url_to_path(tmpdir)

                if not os.path.exists(tmpdir):
                    os.makedirs(tmpdir)

            self.tmpdir = tmpdir
            self.tmpurl = fastr.vfs.path_to_url(self.tmpdir)
            fastr.log.info('Network run tmpdir: {}'.format(self.tmpdir))

            network_file = os.path.join(self.tmpdir, self.NETWORK_DUMP_FILE_NAME)
            self.dumpf(network_file, method='json', indent=2)

            # If needed, send network to PIM for registration
            pim_session = pim_publisher.PimPublisher()
            pim_session.pim_register_run(self)
            self.job_finished_callback = pim_session.pim_update_status
            self.job_status_callback = pim_session.pim_update_status

            try:
                # Set the source and sink data
                for id_, source in self.sourcelist.items():
                    if isinstance(source, ConstantNode):
                        source.set_data()
                    elif id_ in sourcedata:
                        source.set_data(sourcedata[id_])
                    else:
                        raise exceptions.FastrKeyError('Could not find source data for SourceNode {}!'.format(id_))

                    source.update()

                for id_, sink in self.sinklist.items():
                    if id_ not in sinkdata:
                        raise exceptions.FastrKeyError('Could not find sink data for SinkNode {}!'.format(id_))
                    sink.set_data(sinkdata[id_])

                # Create execution objects
                chuncker = DefaultNetworkChunker()
                analyzer = DefaultNetworkAnalyzer()

                # Create a network chuncker to Chunk the Network in executable blocks
                chunks = chuncker.chunck_network(self)
                fastr.log.debug('Found chunks: {}'.format(chunks))

                # Select desired server execution plugin and instantiate
                fastr.log.debug('Selecting {} as executor plugin'.format(fastr.config.execution_plugin))

                # Checlk what execution plugin to use
                if execution_plugin is None:
                    execution_plugin = fastr.config.execution_plugin

                if execution_plugin not in fastr.execution_plugins:
                    message = 'Selected non-existing execution plugin ({}), available plugins: {}'.format(execution_plugin,
                                                                                                          fastr.execution_plugins.keys())
                    fastr.log.error(message)
                    raise exceptions.FastrValueError(message)

                fastr.log.debug('Retrieving execution plugin ({})'.format(execution_plugin))
                execution_interface_type = fastr.execution_plugins[execution_plugin]
                fastr.log.debug('Creating exeuction interface')

                with execution_interface_type(self.job_finished, self.job_finished, self.job_status_callback) as execution_interface:
                    if cluster_queue is not None:
                        execution_interface.default_queue = cluster_queue
                    execution_interface._finished_callback = functools.partial(self.job_finished, execution_interface=execution_interface)
                    execution_interface._cancelled_callback = functools.partial(self.job_finished, execution_interface=execution_interface)

                    for chunk in chunks:
                        if not self.executing:
                            return

                        # Create a network analyzer to create the optimal execution order
                        executionlist = analyzer.analyze_network(self, chunk)

                        joblist = []
                        for node in executionlist:
                            joblist += node.execute()

                            if not self.executing:
                                return

                        fastr.log.debug('Joblist ID: {}'.format([j.id for j in joblist]))

                        # Only try to process chunks that actually have jobs...
                        if len(joblist) > 0:
                            # First queue all jobs that need to run
                            for job in joblist:
                                execution_interface.queue_job(job)

                                if not self.executing:
                                    return

                            fastr.log.info('Waiting for execution to finish...')
                            # Wait in chuncks of two seconds (hopefully this will allow ipython to update more regularly)
                            while len(execution_interface.job_dict) > 0:
                                sys.stdout.flush()
                                if not self.executing:
                                    return
                                time.sleep(1.0)

                        if not self.executing:
                            return

                        fastr.log.info('Chunk execution finished!')

                fastr.log.info("####################################")
                fastr.log.info("#    network execution FINISHED    #")
                fastr.log.info("####################################")
            finally:
                if not self.executing:
                    fastr.log.info("####################################")
                    fastr.log.info("#    network execution ABORTED     #")
                    fastr.log.info("####################################")

                # Make sure to unlock the Network
                fastr.log.debug('Releasing execution lock')
                self.executing = False
                self.execution_lock.set()

                fastr.log.debug('Releasing edit lock')

        fastr.log.info('===== RESULTS =====')
        result = True
        for sink_node, sink_data in sorted(self.sink_results.items()):
            fastr.log.info('---- {} ----'.format(sink_node))
            for sample_id, (job, failed_annotations) in sorted(sink_data.items()):
                fastr.log.info('{}[{}]: {}'.format(sink_node, sample_id, job.status))
                for error in failed_annotations:
                    fastr.log.info('* {e[0]} [{e[1]}]'.format(e=error))
                    result = False
        fastr.log.info('===================')

        sink_data_json = {
            sink_node: {
                str(sample_id): {
                    "status": str(job.status),
                    "errors": [list(str(e) for e in x) for x in failed_annotations],
                    "job_file": job.logfile,
                    "job_url": job.logurl,
                } for sample_id, (job, failed_annotations) in sink_data.items()
            } for sink_node, sink_data in self.sink_results.items()
        }

        sink_result_file = os.path.join(self.tmpdir, self.SINK_DUMP_FILE_NAME)
        with open(sink_result_file, 'w') as fh_out:
            json.dump(sink_data_json, fh_out, indent=2)

        return result

    def abort(self):
        self.executing = False

    def job_finished(self, job, execution_interface):
        """
        Call-back handler for when a job is finished. Will collect the results
        and handle blocking jobs. This function is automatically called when
        the execution plugin finished a job.

        :param job: the job that finished
        :type job: :py:class:`Job <fastr.execution.job.Job>`
        """
        # Find the connected Node
        node_id = job.node_id

        if self.id != job.network_id:
            fastr.log.warning('Jobid network name not matching name of network!'
                              ' ({} vs {})'.format(self.id, job.network_id))
        node = self.nodelist[node_id]

        # Collect earlier failed annotations
        failed_job_annotation = set()
        if not isinstance(node, SourceNode):
            for input_id, input_argument in job.input_arguments.items():
                if not isinstance(input_argument, SampleItem) and isinstance(node, SinkNode):
                    fastr.log.debug('Skipping sink system input: {}'.format(input_argument))
                    continue

                if not isinstance(input_argument, list):
                    input_argument = [input_argument]

                for input_argument_element in input_argument:
                    if str(input_argument_element.id) == '__EMPTY__':
                        continue

                    # Get the annotation from the input
                    fastr.log.debug('Getting annotation from {} --> {}'.format(node.inputs[input_id].get_sourced_outputs(),
                                                                               node.inputs[input_id][input_argument_element.index].failed_annotations))
                    item = node.inputs[input_id][input_argument_element.index]
                    failed_job_annotation.update(item.failed_annotations)

        # Log the status and add annotations if needed
        status_message = "Finished job {} with status {}".format(job.id, job.status)
        fastr.log.info(status_message)

        if job.status not in (JobState.finished, JobState.cancelled):
            if len(job.errors) > 0:
                error = job.errors[0]
                status_message = 'Encountered error: [{e[0]}] {e[1]} ({e[2]}:{e[3]})'.format(e=error)
            else:
                status_message = 'Encountered error in job {}'.format(job.id)

            failed_job_annotation.add((job.id, str(job.status), status_message))

        # Lock Network for editing, make sure not multiple threads will edit network data at the same time
        with self.callback_edit_lock:
            # Data is stored here, need to place it back in the Nodes
            node.set_result(job, failed_job_annotation)

        if self.job_finished_callback is not None:
            #  We will run a callback of which we do not know what it can
            # raise. Since we do not want to callback to crash we use a
            # bare except.
            # pylint: disable=bare-except
            try:
                self.job_finished_callback(job)
            except:
                fastr.log.error('Error in network job finished callback')
                fastr.log.error(traceback.format_exc())
        sys.stdout.flush()

    def is_valid(self):
        def check_object(obj):
            # obj.update()
            status = obj.status
            if not status['valid']:
                if status['messages']:
                    for message in status['messages']:
                        fastr.log.error(message)
                else:
                    fastr.log.error("{} {} is not valid, no message available!".format(type(obj).__name__, obj.id))
            return status['valid']

        valid = True
        for node in self.nodelist.itervalues():
            valid = valid & check_object(node)
        for link in self.linklist.itervalues():
            valid = valid & check_object(link)
        return valid

    def draw_network(self, name="network_layout", img_format='svg', draw_dimension=False):
        """
        Output a dot file and try to convert it to an image file.

        :param str img_format: extension of the image format to convert to
        :return: path of the image created or None if failed
        :rtype: str or None
        """
        allowed_img_formats = ['png', 'svg', 'pdf', 'ps', 'eps', 'gif', 'bmp', 'tif', 'jpg']
        dim_names = ['N', 'M', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L']
        dim_names = dim_names + ["{}{}".format(a, b) for a in dim_names for b in dim_names]
        dim_names.reverse()
        dim = {}

        def lookup_dimname(name):
            if name in dim.keys():
                return dim[name]
            else:
                dim[name] = dim_names.pop()
                return dim[name]

        def format_dimname(dimnames):
            if len(dimnames) > 0:
                return " [{}] ".format("x".join([lookup_dimname(d) for d in dimnames]))
            else:
                return ""

        def print_constant_values(values):
            s = values
            for vi, value in enumerate(values):
                s[vi] = list(value)
                for ii, item in enumerate(value):
                    item = str(item)
                    s[vi][ii] = (item[:12] + '...' + item[-18:]) if len(item) > 33 else item
            return pprint.pformat(s).encode('string_escape')

        if img_format not in allowed_img_formats:
            # Default to svg.
            fastr.log.warning("Image format {} is not supported, falling back to svg".format(img_format))
            img_format = 'svg'

        sourceclr = 'darkolivegreen1'
        nodeclr = 'gray90'
        sinkclr = 'lightskyblue1'
        constantclr = 'plum1'

        dot_file = """digraph G {\n\trankdir = "LR"\n\tsplines=true;\n\tnode [ fontsize="16" shape="record" ];\n\n"""

        # Compile a sourcelist were the constantnodes are left out.
        sourcelist = {x: self.sourcelist[x] for x in self.sourcelist if x not in self.constantlist}

        nodes_in_clusters = []
        for stepid, nodes in self.stepids.items():
            dot_file += """\tsubgraph "cluster_{stepid}" {{\n\t\tlabel="{stepid}";\n\t\tcolor="gray60";\n""".format(stepid=stepid, )
            for node in nodes:
                nodes_in_clusters += [node.id]

                if node.id in sourcelist:
                    clr = sourceclr
                elif node.id in self.constantlist:
                    clr = constantclr
                elif node.id in self.toolnodelist:
                    clr = nodeclr
                elif node.id in self.sinklist:
                    clr = sinkclr
                else:
                    clr = nodeclr

                if node.id in self.constantlist:
                    if draw_dimension:
                        dot_file += '\t\t"{id}" [ style=filled fillcolor={clr} label="<id>{id}|<output>{data} {dim}" ];\n'.format(id=node.id, clr=clr, data=print_constant_values(node.data.values()), dim=format_dimname(node.outputs['output'].dimnames))
                    else:
                        dot_file += '\t\t"{id}" [ style=filled fillcolor={clr} label="<id>{id}|<output>{data}" ];\n'.format(id=node.id, clr=clr, data=print_constant_values(node.data.values()))
                else:
                    if draw_dimension:
                        format_inputs = '|'.join(['<i_{k}>{dim}{k}'.format(k=k, dim=format_dimname(v.dimnames)) for k, v in node.inputs.items()])
                        format_outputs = '|'.join(['<o_{k}>{k}{dim}'.format(k=k, dim=(format_dimname(v.dimnames) if len(v.listeners) > 0 else "")) for k, v in node.outputs.items()])
                    else:
                        format_inputs = '|'.join(['<i_{i}>{i}'.format(i=i) for i in node.inputs.keys()])
                        format_outputs = '|'.join(['<o_{i}>{i}'.format(i=i) for i in node.outputs.keys()])
                    dot_file += '\t\t"{id}" [ style=filled fillcolor={clr} label="<id>{id}|{{{{{inputs}}}|{{{outputs}}}}}" ];\n'.format(id=node.id, clr=clr, inputs=format_inputs, outputs=format_outputs)
            dot_file += "\t}\n\n"

        for clr, lst in zip([sourceclr, constantclr, nodeclr, sinkclr], [sourcelist, self.constantlist, self.toolnodelist, self.sinklist]):
            for id_, node in lst.items():
                if id_ not in nodes_in_clusters:
                    if node.id in self.constantlist:
                        if draw_dimension:
                            dot_file += '\t"{id}" [ style=filled fillcolor={clr} label="<o_id>{id}|<output>{data} {dim}" ];\n'.format(id=node.id, clr=clr, data=print_constant_values(node.data.values()), dim=format_dimname(node.outputs['output'].dimnames))
                        else:
                            dot_file += '\t"{id}" [ style=filled fillcolor={clr} label="<o_id>{id}|<output>{data}" ];\n'.format(id=node.id, clr=clr, data=print_constant_values(node.data.values()))
                    else:
                        if draw_dimension:
                            format_inputs = '|'.join(['<i_{k}>{dim}{k}'.format(k=k, dim=format_dimname(v.dimnames)) for k, v in node.inputs.items()])
                            format_outputs = '|'.join(['<o_{k}>{k}{dim}'.format(k=k, dim=(format_dimname(v.dimnames) if len(v.listeners) > 0 else "")) for k, v in node.outputs.items()])
                        else:
                            format_inputs = '|'.join(['<i_{i}>{i}'.format(i=i) for i in node.inputs.keys()])
                            format_outputs = '|'.join(['<o_{i}>{i}'.format(i=i) for i in node.outputs.keys()])
                        dot_file += '\t"{}" [ style=filled fillcolor={} label="<id>{}|{{{{{inputs}}}|{{{outputs}}}}}" ];\n'.format(node.id, clr, node.id, inputs=format_inputs, outputs=format_outputs)
        for link in self.linklist.values():
            source_node_id = link.source.node.id
            source_output_id = link.source.id
            target_node_id = link.target.node.id
            target_input_id = link.target.id
            edge_properties = " ["
            if link.collapse and not link.expand:
                edge_properties += " color=blue weight=3 penwidth=3"
            elif link.expand and not link.collapse:
                edge_properties += " color=red weight=3 penwidth=3"
            elif link.expand and link.collapse:
                edge_properties += " color=purple weight=3 penwidth=3"
            edge_properties += "]"
            dot_file += '\t"{}":o_{} -> "{}":i_{}{props};\n'.format(source_node_id, source_output_id, target_node_id, target_input_id, props=edge_properties)
        dot_file += "}"

        dot_file_path = '{}.dot'.format(name)
        if os.path.isfile(dot_file_path):
            os.remove(dot_file_path)

        image_file_path = '{}.{}'.format(name, img_format)
        if os.path.exists(image_file_path):
            os.remove(image_file_path)
        with open(dot_file_path, 'wb') as output_file:
            output_file.write(dot_file)
        filename = os.path.join(os.getcwd(), '{}.dot'.format(name))
        fastr.log.debug("%s file created.", filename)
        try:
            proc = subprocess.Popen(['dot', '-T{}'.format(img_format), '-o{}'.format(image_file_path), dot_file_path],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT)
            stdout, stderr = proc.communicate()
            fastr.log.debug('Subprocess call to do finished with stdout: {}, stderr: {}'.format(stdout, stderr))

            if not os.path.exists(image_file_path):
                fastr.log.error('Network draw failed the graphviz coversion:\n{}'.format(stdout))
            fastr.log.debug("converted to %s", image_file_path)
            return image_file_path
        except OSError:
            fastr.log.warning("Cannot convert %s to an svg image. Please put dot (from GraphViz) in your PATH.", filename)
            return None
